'''
    Times.

    Utilties for times. Great for working with time series.

    Todo: Handle timezones better. See parse_timestamp(timezone=...),

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-29

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import calendar
import re
import time
# avoid conflict with 'timezone' param
from datetime import timezone as datetime_timezone
from datetime import date, datetime, timedelta

try:
    import pytz
    from pytz.exceptions import UnknownTimeZoneError
except ImportError:
    PYTZ_AVAILABLE = False
else:
    PYTZ_AVAILABLE = True

from denova.net.web_log_parser import LogLine
from denova.python.log import get_log
from denova.python.format import s_if_plural

# map month abbrevs to numeric equivalent
MONTH_MAP = {'Jan': 1,
             'Feb': 2,
             'Mar': 3,
             'Apr': 4,
             'May': 5,
             'Jun': 6,
             'Jul': 7,
             'Aug': 8,
             'Sep': 9,
             'Oct': 10,
             'Nov': 11,
             'Dec': 12}

# ISO 8901 datetime format with microseconds and timezone
ISO_DATETIME_RE = r'(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)[ T](?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)\.(?P<microsecond>\d*)(?P<timezone>.*)'
TIMEZONE_RE = re.compile(r'(?P<name>[a-zA-Z]{3})? *?((?P<sign>[+|-])(?P<hour>\d\d):(?P<minute>\d\d))?')

seconds_in_minute = 60
seconds_in_hour = 60 * seconds_in_minute
seconds_in_day = 24 * seconds_in_hour
microseconds_in_second = 1000000

# we don't define one_month, one_year, etc.
# because the exact definition varies from app to app
one_year = timedelta(days=365) # ignores leap year
one_week = timedelta(days=7)
one_day = timedelta(days=1)
one_hour = timedelta(hours=1)
one_minute = timedelta(minutes=1)
one_second = timedelta(seconds=1)
one_millisecond = timedelta(milliseconds=1)
one_microsecond = timedelta(microseconds=1)
no_time = timedelta(0)

far_past = datetime.min
far_future = datetime.max
# indicate all dates and times
anytime = far_future - one_microsecond

log = get_log()
_compiled_timestamps = []

# some constants are defined after we define functions they need

def now(utc=True):
    ''' Get the current datetime.

        If utc=True, return UTC date/time.

        Returns a datetime object.

        >>> dt = now()
        >>> repr(dt.tzinfo)
        'datetime.timezone.utc'
    '''

    if utc:
        tzinfo = datetime_timezone.utc
    else:
        tzinfo = None

    if utc:
        dt = datetime.utcnow().replace(tzinfo=tzinfo)
    else:
        dt = datetime.now()
    return dt

def get_short_now(utc=True):
    '''Get datetime up to the minute, not the second or millisecond.

        >>> get_short_now().second == 0 and get_short_now().microsecond == 0
        True
        >>> get_short_now(utc=True).second == 0 and get_short_now(utc=True).microsecond == 0
        True
    '''

    if utc:
        tzinfo = datetime_timezone.utc
    else:
        tzinfo = None

    time_now = now(utc=utc)
    return datetime(time_now.year, time_now.month, time_now.day,
                    time_now.hour, time_now.minute,
                    tzinfo=tzinfo)

def timestamp(when=None, microseconds=True):
    ''' Return timestamp as a string in a standard format. Time zone is UTC.

        'when' is a datetime or timestamp string. Defaults to now.

        If 'microseconds' is True microseconds are included. Defaults to True.

        >>> re.match(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d*?\+00:00$', timestamp()) is not None
        True
        >>> re.match(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$', timestamp(microseconds=False)) is not None
        True
    '''

    if when:
        if isinstance(when, str):
            when = parse_timestamp(when)
        elif not isinstance(when, datetime):
            raise ValueError("'when' must be a datetime, date/time string, or None")
    else:
        when = now()

    # datetime.str() returns ISO format
    formatted_time = when.isoformat(sep=' ')

    if not microseconds:
        i = formatted_time.rfind('.')
        if i > 0:
            formatted_time = formatted_time[:i]

    #log('formatted_time: {}'.format(formatted_time))

    return formatted_time

def parse_timestamp(timestr, startswith=False, default_year=None, timezone=None):
    ''' Parse a string timestamp. Return datetime.

        Return first timestamp found.

        If no valid timestamp found, raises ValueError.

        If startswith is set to True, only match a timestamp at the
        beginning of timestr. The default is False, which means match
        a timestamp anywhere in timestr.

        Year defaults to current year, which won't be right if, for example,
        a process is running when a new year starts.. Microseconds default
        is zero.

        'timezone' is a timezone string.
        Examples are 'UTC', 'EST' and 'Europe/Amsterdam'.
        To keep django happy, parse_timestamp() defaults to UTC.
        OUTDATED:
        It is used only if the timestr is "naive", meaning has no timezone.
        If timestr does not have a timezone and 'timezone' is specified,
        then 'timezone' is passed to pytz.timezone(), and
        parse_timestamp's returned  datetime has the timezone from
        pytz.timezone(). If pytz is not installed, the returned datetime
        is "naive", meaning it does not have a timezone.

        If timestr does not have a timezone and
        'timezone' is not specified, the returned datetime is "naive",
        meaning it does not have a timezone. It's a good idea to always
        specify 'timezone='.

        parse_timestamp() handles multiple formats.

        >>> # RFC 2616 required formats
        >>> #     RFC 822, updated by RFC 1123
        >>> ts = parse_timestamp('Sun, 06 Nov 1994 08:49:37 GMT')
        >>> ts.year
        1994
        >>> #     RFC 850, obsoleted by RFC 1036
        >>> ts = parse_timestamp('Sunday, 06-Nov-94 08:49:37 GMT')
        >>> ts.year
        94
        >>> #     ANSI C's asctime() format
        >>> ts = parse_timestamp('Sun Nov  6 08:49:37 1994')
        >>> ts.year
        1994
        >>> ts = parse_timestamp('Mon, 27 Apr 2020 15:55:56 GMT')
        >>> ts.day
        27

        >>> # no year
        >>> ts = parse_timestamp('Oct 28 11:06:55.000')
        >>> ts.year > 0
        True
        >>> ts.month
        10
        >>> ts.day
        28
        >>> ts.hour
        11
        >>> ts.minute
        6
        >>> ts.second
        55
        >>> ts.microsecond
        0

        >>> # no microseconds
        >>> parse_timestamp('Tue Jan 15 14:49:13 2000')
        datetime.datetime(2000, 1, 15, 14, 49, 13)

        >>> # no year and no microseconds
        >>> ts = parse_timestamp('Oct 28 11:06:55')
        >>> ts.year > 0
        True
        >>> ts.month
        10
        >>> ts.day
        28
        >>> ts.hour
        11
        >>> ts.minute
        6
        >>> ts.second
        55
        >>> ts.microsecond
        0

        # no year, single digit day, no microseconds
        >>> ts = parse_timestamp('Apr  9 18:17:01')
        >>> ts.year > 0
        True
        >>> ts.month
        4
        >>> ts.day
        9
        >>> ts.hour
        18
        >>> ts.minute
        17
        >>> ts.second
        1
        >>> ts.microsecond
        0

        >>> repr(parse_timestamp('1970-01-18T20:03:11.282Z'))
        'datetime.datetime(1970, 1, 18, 20, 3, 11, 282, tzinfo=datetime.timezone.utc)'

        >>> repr(parse_timestamp('2019-10-17 19:46:30.574Z'))
        'datetime.datetime(2019, 10, 17, 19, 46, 30, 574, tzinfo=datetime.timezone.utc)'

        >>> repr(parse_timestamp('17/Oct/2019:09:35:52 +0000'))
        'datetime.datetime(2019, 10, 17, 9, 35, 52, tzinfo=datetime.timezone.utc)'

        We're not handling this date string because
        different parts of the world interrupt this date differently.
        repr(parse_timestamp('02:07:36 05/08/03 EDT'))
        'datetime.datetime(5, 8, 3, 2, 7, 36)'

        >>> repr(parse_timestamp('2019-10-26 11:38:05.000711', timezone='UTC'))
        'datetime.datetime(2019, 10, 26, 11, 38, 5, 711, tzinfo=datetime.timezone.utc)'

        >>> repr(parse_timestamp('2019/10/17 12:40:36'))
        'datetime.datetime(2019, 10, 17, 12, 40, 36)'

        >>> repr(parse_timestamp('2019-10-23T17:55:00'))
        'datetime.datetime(2019, 10, 23, 17, 55)'

        >>> repr(parse_timestamp('2019-10-23T17:55:00UTC-07:30'))
        'datetime.datetime(2019, 10, 23, 17, 55, tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=59400)))'

        >>> repr(parse_timestamp('2019-04-11T12:49:00 UTC+03:00'))
        'datetime.datetime(2019, 4, 11, 12, 49, tzinfo=datetime.timezone(datetime.timedelta(seconds=10800)))'

        >>> repr(parse_timestamp('Oct 28 09:54:48'))
        'datetime.datetime(2020, 10, 28, 9, 54, 48)'

        >>> repr(parse_timestamp(b'Oct 28 09:54:48'))
        'datetime.datetime(2020, 10, 28, 9, 54, 48)'

        >>> if PYTZ_AVAILABLE:
        ...     log.debug('test pytz')
        ...     tz = pytz.timezone('UTC')
        ...     tz.zone
        'UTC'

        >>> if PYTZ_AVAILABLE:
        ...     tz = pytz.timezone('EST')
        ...     tz.zone
        'EST'

        >>> if PYTZ_AVAILABLE:
        ...     # pytz can't handle Daylight Savings
        ...     try:
        ...         pytz.timezone('EDT')
        ...     except pytz.exceptions.UnknownTimeZoneError:
        ...         pass

        >>> if PYTZ_AVAILABLE:
        ...     tz = pytz.timezone('Europe/Amsterdam')
        ...     tz.zone
        'Europe/Amsterdam'
    '''

    def parse_timezone(timezone_str):
        ''' Parse the timezone as an offset from UTC.

            Try the pytz module first.

            Returns tzinfo or None.
        '''

        #log.debug('timezone_str: {}'.format(timezone_str))
        tzinfo = None

        if PYTZ_AVAILABLE:
            try:
                tzinfo = pytz.timezone(timezone_str)
            except UnknownTimeZoneError:
                #log.debug('timezone unknown to pytz: {}'.format(timezone_str))
                pass

        if not tzinfo:
            match = TIMEZONE_RE.search(timezone_str)
            if match:
                mdict = match.groupdict()
                #log.debug('timezone dict: {}'.format(mdict))

                # name must be UTC; if it's not and pytz wasn't able to
                # handle the conversion, then we'll ignore the timezone
                if 'name' in mdict:
                    if mdict['name']:
                        timezone_is_utc = mdict['name'].upper() == 'UTC'
                    else:
                        timezone_is_utc = True
                else:
                    timezone_is_utc = True

                if timezone_is_utc:
                    # timedelta() requires hour/minute
                    if ('sign' in mdict) and ('hour' in mdict) and ('minute' in mdict):

                        tz_sign = mdict['sign']
                        tz_hours = int(mdict['hour'])
                        tz_minutes = int(mdict['minute'])

                        offset = timedelta(hours=tz_hours, minutes=tz_minutes)
                        if tz_sign == '-':
                            offset = -offset
                        tzinfo = datetime_timezone(offset)
            else:
                log.debug(f'no timezone parsed. TIMEZONE_RE did not match. timestr: {timestr}')

        return tzinfo

    def get_timestamp(mdict, timezone):
        """ Timestamp regular expressions

            Match keywords must be the names of datetime.datetime parameters.

            The order of patterns is important. The patterns are tried in order.
            The first one to match is accepted. That means you need to
            list more complete ones first. For example, you should specify
            the tor log format before the syslog format. They are the same
            except the tor log format includes microseconds. So you want to
            match on microseconds if they are available, but still match
            on a different pattern when microseconds aren't available.

            Web log format is a special case handled in the code.
        """

        DATETIME_KEYWORDS = ['year', 'month', 'day',
                             'hour', 'minute', 'second', 'microsecond',
                             'tzinfo']

        #log.debug('mdict {}'.format(mdict)) # DEBUG

        kwargs = {}
        for key in mdict:
            if key == 'timezone':
                value = mdict[key]
                kwargs[key] = value.strip()

            elif key == 'month':
                month = mdict[key]
                try:
                    month = int(month)
                except ValueError:
                    # try to map a text month to an int
                    month = MONTH_MAP[month]
                except AttributeError:
                    raise ValueError('month must be an integer or month string')
                else:
                    if month < 1 or month > 12:
                        raise ValueError('month must be an integer in the range 1 to 12, or month string')
                kwargs[key] = month

            # ignore 'weekday', etc.
            elif key in DATETIME_KEYWORDS:
                value = int(mdict[key])
                kwargs[key] = value

        # datetime() requires year/month/day

        # not all timestamps have a year
        if 'year' not in kwargs:
            # we default to this year, which won't be right if this
            # process runs into a new year.
            kwargs['year'] = default_year or now().year

        #log.debug('get tzinfo. timezone={}'.format(timezone)) # DEBUG
        ''' We want to set tzinfo from the timestr if possible.
            Second choice is the timezone= passed to parse_timestamp().
            Last choice is default to a "naive" datetime, with no tzinfo.
        '''
        tzinfo = None

        # if timestr has a timezone, override any timezone param
        if 'timezone' in kwargs:
            tz = kwargs['timezone']
            del kwargs['timezone']
            #log.debug('kwargs[timezone]: {}'.format(tz)) # DEBUG
            if tz:
                timezone = tz
                #log.debug('set timezone from kwargs: {}'.format(timezone)) # DEBUG

        if timezone:
            if timezone.upper() in ['UTC','GMT', 'Z']:
                tzinfo = datetime_timezone.utc
            else:
                tzinfo = parse_timezone(timezone)

        if tzinfo:
            kwargs['tzinfo'] = tzinfo
            #log.debug('tzinfo: {}'.format(tzinfo))
        #else:
            #log.warning('no tzinfo') # DEBUG

        try:
            timestamp = datetime(**kwargs)
        except ValueError:
            log(f'bad datetime values: {repr(kwargs)}')
            log.exception()

        return timestamp

    RAW_TIMESTAMP_RES = [# iso datetime, which is datetime default
         ISO_DATETIME_RE,

         # short iso datetime with timezone
         r'(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)[ T](?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)(?P<timezone>.*)',

         # short iso datetime
         r'(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)[ T](?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)',

         # nginx error log
         # Example: 2019/10/17 12:40:36
         r'(?P<year>\d+)/(?P<month>\d+)/(?P<day>\d+) (?P<hour>\d+):(?P<minute>\d+):(?P<second>\d+)',

         # RFC 850, obsoleted by RFC 1036, Example: Sunday, 06-Nov-94 08:49:37 GMT
         r'(?P<weekday>[A-Za-z]+),\s+(?P<day>\d+)-(?P<month>[A-Za-z]+)-(?P<year>\d+)\s+(?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d)\s+(?P<timezone>.*)',

         # Example: Mon, 27 Apr 2020 15:55:56 GMT
         r'(?P<weekday>[A-Za-z]+),\s(?P<day>\d+)\s+(?P<month>[A-Za-z]+)\s+(?P<year>\d+)\s+(?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d)\s+(?P<timezone>.*)',

         # Example: Tue Jan 15 14:49:13 2019
         r'(?P<weekday>[A-Za-z]+)\s+(?P<month>[A-Za-z]+)\s+(?P<day>\d+)\s+(?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d)\s+(?P<year>\d+)',

         # tor log
         # Example: Oct 28 11:06:55.000
         r'(?P<month>[A-Za-z]+)\s+(?P<day>\d\d)[ T](?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d)\.(?P<microsecond>\d\d\d)',

         # syslog
         # Example: Oct 28 11:06:55
         # same as tor log, but without microseconds
         r'(?P<month>[A-Za-z]+)\s+(?P<day>\d+)[ T](?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d)',

         # Example: '02:07:36 05/08/03 EDT'
         # Different parts of the world interpret this differently so we're going to ignore it
         #r'(?P<hour>\d+):(?P<minute>\d\d):(?P<second>\d\d) (?P<year>\d\d)/(?P<month>\d+)/(?P<day>\d+) (?P<timezone>.*)',
        ]

    #log.debug('timestr: {}'.format(repr(timestr))) # DEBUG

    if not _compiled_timestamps:
        # anchor timestamp formats to start of line and compile them
        for raw_re in RAW_TIMESTAMP_RES:

            # if startswith, only match timestamps at the start of the line
            if startswith and not raw_re.startswith(r'^'):
                compiled_re = re.compile(r'^' + raw_re)
            else:
                compiled_re = re.compile(raw_re)

            _compiled_timestamps.append(compiled_re)

    timestamp = None

    if type(timestr) is datetime:
        # save some callers from having to check the type
        timestamp = timestr

    elif type(timestr) is bytes:
        timestr = str(timestr)

    if not timestamp:
        for timestamp_re in _compiled_timestamps:
            if not timestamp:
                try:
                    # sometimes we get "must be str or a bytes-like object"
                    #log.debug('type(timestamp_re): {}'.format(type(timestamp_re))) # DEBUG
                    #log.debug('type(timestr): {}'.format(type(timestr))) # DEBUG
                    match = timestamp_re.search(timestr)
                    if match:
                        #log.debug('timestamp_re matched: {}'.format(timestamp_re)) # DEBUG
                        timestamp = get_timestamp(match.groupdict(), timezone)

                except Exception as e:
                    log.debug(f'unable to parse timestamp: {timestr}')
                    from denova.python.utils import format_exception
                    log.debug(format_exception(e))

    if not timestamp:
        # web log format is a special case (not at start of timestr), so try it last
        try:
            timestamp = LogLine.get_timestamp(timestr)

        except ValueError:
            pass

        except Exception as e:
            log.debug(f'unable to parse timestamp as web log line: {timestr}')
            from denova.python.utils import format_exception
            log.debug(format_exception(e))

    if not timestamp:
        raise ValueError(timestr)

    if timezone:
        # we should have a tzinfo attribute
        try:
            if not timestamp.tzinfo:
                raise Exception(f'tzinfo is None. timestr: {timestr}, timezone={timezone}')
        except AttributeError:
            raise Exception(f'tzinfo not set. timestr: {timestr}, timezone={timezone}')

    return timestamp

def format_time(d):
    ''' Return the item in the ISO 8601 format.

        If d is not str, int, or float, or datetime, throws ValueError.

        A string is any datetime string that parse_timestamp() can parse.
        An int or float is seconds since the epoch.

        >>> format_time(datetime(2012, 7, 6, 17, 57, 12, 4))
        '2012-07-06 17:57:12.000004'
        >>> format_time('2019-10-17 19:46:30.574Z')
        '2019-10-17 19:46:30.000574+00:00'
        >>> format_time('17/Oct/2019:09:35:52 +0000')
        '2019-10-17 09:35:52+00:00'
        >>> format_time('2019-10-23T17:55:00')
        '2019-10-23 17:55:00+00:00'
        >>> format_time(1571760736.9401658)
        '2019-10-22 16:12:16.940166+00:00'
        >>> format_time(1571760736)
        '2019-10-22 16:12:16+00:00'
        >>> format_time(1571760737)
        '2019-10-22 16:12:17+00:00'

        We're not handling this date string because
        different parts of the world interrupt this date differently.
        format_time('02:07:36 05/08/03 EDT')
        '0005-08-03 02:07:36'
    '''

    if d is None:
        raise ValueError

    elif isinstance(d, datetime):
        dt = d

    elif isinstance(d, str):
        dt = parse_timestamp(d, timezone='UTC')

    elif isinstance(d, (float, int)):
        dt = seconds_to_datetime(d)

    else:
        raise ValueError

    return dt.isoformat(sep=' ')

def format_date(date=None):
    ''' Return the date in a standard format.

        If date is not specified or the date is None, formats today's date.
        Time zone is UTC.

        >>> format_date((2012, 7, 6, 17, 57, 12, 4, 188, 0))
        '2012-07-06'
        >>> re.match(r'^\d{4}-\d{2}-\d{2}$', format_date()) is not None
        True
    '''

    if not date:
        date = time.gmtime()
    return time.strftime('%Y-%m-%d', date)

def seconds_human_readable(seconds):
    '''
        Formats seconds in a human readable form.

        If seconds is less than 0, then return None.
        seconds: 115619.434662
         days: 1.0
         minutes: 419
         hours: 32.0
         status: 32 hours, 419 minutes
        seconds: 29994.513938
         days: 0.0
         minutes: 1194
         hours: 8.0
         status: 8 hours, 1194 minutes


        >>> seconds_in_week = seconds_in_day * 7
        >>> seconds_in_year = seconds_in_week * 52
        >>> current = datetime.utcnow()
        >>> hour_ago = current - timedelta(minutes=48, seconds=20)
        >>> seconds = (current - hour_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '48 minutes'
        >>> hour_ago = current - timedelta(minutes=60)
        >>> seconds = (current - hour_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '1 hour'
        >>> hours_ago = current - timedelta(hours=5, minutes=6)
        >>> seconds = (current - hours_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '5 hours, 6 minutes'
        >>> two_days_ago = current - timedelta(days=2)
        >>> seconds = (current - two_days_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '2 days'
        >>> five_plus_weeks_ago = current - timedelta(days=37)
        >>> seconds = (current - five_plus_weeks_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '5 weeks, 2 days'
        >>> eight_plus_years_ago = current - timedelta(days=2921)
        >>> seconds = (current - eight_plus_years_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '8 years, 1 week'
        >>> three_years_ago = current - timedelta(days=1095)
        >>> seconds = (current - three_years_ago).total_seconds()
        >>> seconds_human_readable(seconds)
        '3 years'
    '''

    def format_unit_label(label, units):
        ''' Format the label. '''
        if units == 1:
            status = f'{units} {label}'
        else:
            status = f'{units} {label}s'

        return status

    status = None

    seconds_in_week = seconds_in_day * 7
    seconds_in_year = seconds_in_week * 52

    years = seconds // seconds_in_year
    if years > 0:
        years_status = format_unit_label('year', int(years))
        weeks = (seconds % seconds_in_year) // seconds_in_week
        if weeks > 0:
            weeks_status = format_unit_label('week', int(weeks))
            status = f'{years_status}, {weeks_status}'
        else:
            status = years_status

    if status is None:
        weeks = seconds // seconds_in_week
        if weeks > 0:
            weeks_status = format_unit_label('week', int(weeks))
            days = (seconds % seconds_in_week) // seconds_in_day
            if days > 0:
                days_status = format_unit_label('day', int(days))
                status = f'{weeks_status}, {days_status}'
            else:
                status = weeks_status

    if status is None:
        days = seconds // seconds_in_day
        if days > 0:
            days_status = format_unit_label('day', int(days))
            hours = (seconds % seconds_in_day) // seconds_in_hour
            if hours > 0:
                hours_status = format_unit_label('hour', int(hours))
                status = f'{days_status}, {hours_status}'
            else:
                status = days_status

    if status is None:
        hours = seconds // seconds_in_hour
        if hours > 0:
            hours_status = format_unit_label('hour', int(hours))
            minutes = (seconds % seconds_in_hour) // seconds_in_minute
            if minutes > 0:
                minutes_status = format_unit_label('minute', int(minutes))
                status = f'{hours_status}, {minutes_status}'
            else:
                status = hours_status

    if status is None:
        minutes = seconds // seconds_in_minute
        if minutes > 1:
            status = format_unit_label('minute', int(minutes))
        else:
            secs = seconds % seconds_in_minute
            if secs > 0:
                status = format_unit_label('second', int(secs))

    return status

def timedelta_to_human_readable(td, verbose=True):
    ''' Formats a timedelta in a human readable form.

        If total time is less than a second then shows milliseconds
        instead of microseconds, else rounds to the nearest second.

        >>> timedelta_to_human_readable(timedelta(days=1, seconds=123, minutes=4, hours=26))
        '2 days, 2 hours, 6 minutes, 3 seconds'
        >>> timedelta_to_human_readable(timedelta(seconds=123))
        '2 minutes, 3 seconds'
        >>> timedelta_to_human_readable(timedelta(seconds=65))
        '1 minute, 5 seconds'
        >>> timedelta_to_human_readable(timedelta(milliseconds=85))
        '85 ms'
        >>> timedelta_to_human_readable(timedelta(days=1, seconds=123, minutes=4, hours=26), verbose=False)
        '2days, 2hrs, 6mins, 3secs'
        >>> timedelta_to_human_readable(timedelta(seconds=123), verbose=False)
        '2mins, 3secs'
        >>> timedelta_to_human_readable(timedelta(seconds=65), verbose=False)
        '1min, 5secs'
        >>> timedelta_to_human_readable(timedelta(milliseconds=85), verbose=False)
        '85 ms'
    '''

    tdString = ''

    if td.days or td.seconds:

        # days
        if td.days:
            tdString = f'{td.days} day{s_if_plural(td.days)}'

        # round seconds
        seconds = td.seconds
        if (td.microseconds * 2) >= td.max.microseconds:
            seconds = seconds + 1

        # hours
        hours = seconds // seconds_in_hour
        if hours:
            if tdString:
                tdString = tdString + ', '
            tdString = tdString + f'{hours} hour{s_if_plural(hours)}'

        # minutes
        secondsLeft = seconds - (hours * seconds_in_hour)
        if secondsLeft:
            minutes = secondsLeft // seconds_in_minute
            if minutes:
                if tdString:
                    tdString = tdString + ', '
                tdString = tdString + f'{minutes} minute{s_if_plural(minutes)}'
                secondsLeft = secondsLeft - (minutes * seconds_in_minute)

        # seconds
        if secondsLeft:
            if tdString:
                tdString = tdString + ', '
            tdString = tdString + f'{secondsLeft} second{s_if_plural(secondsLeft)}'

    else:
        milliseconds = (td.microseconds + 1) / 1000
        tdString = f'{milliseconds} ms'

    if not verbose:
        m = re.match('.*( day)', tdString)
        if m:
            tdString = tdString.replace(m.group(1), 'day')

        m = re.match('.*( hour)', tdString)
        if m:
            tdString = tdString.replace(m.group(1), 'hr')

        m = re.match('.*( second)', tdString)
        if m:
            tdString = tdString.replace(m.group(1), 'sec')

        m = re.match('.*( minute)', tdString)
        if m:
            tdString = tdString.replace(m.group(1), 'min')

    return tdString


def get_short_date_time(date_time):
    '''Format the date-time without seconds.

        >>> get_short_date_time(datetime(2012, 6, 1, 12, 30, 0))
        '2012-06-01 12:30'
        >>> get_short_date_time(datetime(2012, 6, 1, 12, 30, 41))
        '2012-06-01 12:30'
        >>> get_short_date_time(datetime(2012, 6, 1, 12, 30, 0, 0))
        '2012-06-01 12:30'
        >>> get_short_date_time(None)
        ''
    '''

    if date_time:
        new_date_time = date_time.isoformat(sep=' ')
        try:
            m = re.match(r'.*?(\d+\:\d+\:\d+).*', new_date_time)
            if m:
                current_time = m.group(1)
                index = current_time.rfind(':')
                new_date_time = new_date_time.replace(m.group(1), current_time[:index])
        except Exception:
            pass

    else:
        new_date_time = ''

    return new_date_time

def datetime_to_date(dt):
    ''' Converts a datetime to a date. If dt is a date, returns a copy.

        >>> datetime_to_date(datetime(2012, 6, 1, 12, 30, 0))
        datetime.date(2012, 6, 1)
        >>> datetime_to_date(datetime(2012, 6, 1, 12, 30, 0, 0))
        datetime.date(2012, 6, 1)
    '''
    return date(dt.year, dt.month, dt.day)

def date_to_datetime(d):
    ''' Converts a date or datetime to a datetime at the beginning of the day.

        >>> date_to_datetime(datetime(2012, 6, 1, 1, 39))
        datetime.datetime(2012, 6, 1, 0, 0, tzinfo=datetime.timezone.utc)
        >>> date_to_datetime(date(2012, 6, 1))
        datetime.datetime(2012, 6, 1, 0, 0, tzinfo=datetime.timezone.utc)
    '''

    return datetime(d.year, d.month, d.day, tzinfo=datetime_timezone.utc)

def timedelta_to_days(td):
    ''' Converts timedelta to floating point days.

        >>> timedelta_to_days(timedelta(seconds=864000))
        10.0
        >>> timedelta_to_days(timedelta(days=4, seconds=43200))
        4.5
    '''

    # !!!!! python 3
    return timedelta_to_seconds(td) / seconds_in_day

def timedelta_to_seconds(td):
    ''' Converts timedelta to floating point seconds.

        >>> timedelta_to_seconds(timedelta(seconds=864000))
        864000.0
        >>> timedelta_to_seconds(timedelta(days=4, seconds=43200))
        388800.0
    '''

    # internally timedelta only stores days, seconds, and microseconds
    # !!!!! python 3
    return (
        (td.days * seconds_in_day) +
        td.seconds +
        ((1.0 * td.microseconds) / microseconds_in_second))

def datetime_to_seconds(dt):
    ''' Converts datetime to floating point seconds since the epoch.

        >>> dt =datetime(2012, 6, 1, 12, 30, 0)
        >>> secs = calendar.timegm(dt.timetuple())
        >>> secs
        1338553800
        >>> dt_secs = datetime_to_seconds(dt)
        >>> dt_secs == secs
        True
    '''

    return calendar.timegm(dt.timetuple())

def seconds_to_datetime(seconds):
    ''' Convert seconds since the epoch to a datetime.

        Convenience function.

        >>> seconds_to_datetime(864000)
        datetime.datetime(1970, 1, 11, 0, 0, tzinfo=datetime.timezone.utc)
    '''

    utc = datetime_timezone.utc
    return datetime.fromtimestamp(float(seconds), tz=utc)

def one_month_before(date):
    ''' Returns one month before.

        Accepts a date or datetime. Returns a datetime.

        >>> one_month_before(date(2012, 6, 1))
        datetime.datetime(2012, 5, 1, 0, 0)
        >>> one_month_before(date(2012, 3, 30))
        datetime.datetime(2012, 2, 29, 0, 0)
        >>> one_month_before(date(2012, 5, 31))
        datetime.datetime(2012, 4, 30, 0, 0)
    '''

    if date.month > 1:
        last_year = date.year
        last_month = date.month - 1
    else:
        last_year = date.year - 1
        last_month = 12

    current_date_range = calendar.monthrange(date.year, date.month)
    last_date_range = calendar.monthrange(last_year, last_month)
    if date.day == current_date_range[1]:
        last_day = last_date_range[1]
    else:
        if date.day > last_date_range[1]:
            last_day = last_date_range[1]
        else:
            last_day = date.day

    earlier = datetime(last_year, last_month, last_day)
    return earlier

def start_of_day(d):
    ''' Returns datetime with no hours, minutes, seconds, or microseconds.

        Accepts a date or datetime. Returns a datetime.

        >>> start_of_day(date(2012, 6, 1))
        datetime.datetime(2012, 6, 1, 0, 0, tzinfo=datetime.timezone.utc)
        >>> start_of_day(datetime(2012, 3, 30, 11, 27))
        datetime.datetime(2012, 3, 30, 0, 0, tzinfo=datetime.timezone.utc)
    '''

    return date_to_datetime(datetime_to_date(d))

def end_of_day(d):
    ''' Returns the latest datetime for the day.

        Accepts a date or datetime. Returns a datetime.

        >>> end_of_day(date(2012, 6, 1))
        datetime.datetime(2012, 6, 1, 23, 59, 59, 999999, tzinfo=datetime.timezone.utc)
        >>> end_of_day(datetime(2012, 3, 30, 11, 27))
        datetime.datetime(2012, 3, 30, 23, 59, 59, 999999, tzinfo=datetime.timezone.utc)
    '''

    return start_of_day(d) + one_day - one_microsecond

def date_range(start, end, lei_convention=False):
    ''' Generates every date in the range from start to end, inclusive.

        To follow the endpoint convention [a, b) and exclude the
        right endpoint, i.e. include the first element and exclude the last,
        set lei_convention=True. See http://mathworld.wolfram.com/Interval.html

        If start is later than end, returns dates in reverse chronological
        order.

        Accepts dates or datetimes. Yields dates.

        >>> list(date_range(date(2012, 6, 1), date(2012, 6, 2)))
        [datetime.date(2012, 6, 1), datetime.date(2012, 6, 2)]
        >>> list(date_range(date(2012, 6, 2), date(2012, 6, 1)))
        [datetime.date(2012, 6, 2), datetime.date(2012, 6, 1)]
        >>> list(date_range(date(2012, 6, 1), date(2012, 6, 1)))
        [datetime.date(2012, 6, 1)]
        >>> list(date_range(date(2012, 6, 1), date(2012, 6, 2), lei_convention=True))
        [datetime.date(2012, 6, 1)]
        >>> list(date_range(date(2012, 6, 2), date(2012, 6, 1), lei_convention=True))
        [datetime.date(2012, 6, 2)]
        >>> list(date_range(datetime(2012, 6, 1, 21, 3), datetime(2012, 6, 2, 0, 0)))
        [datetime.date(2012, 6, 1), datetime.date(2012, 6, 2)]
        >>> list(date_range(datetime(2012, 6, 1, 0, 0), datetime(2012, 6, 2, 23, 59, 59, 999999)))
        [datetime.date(2012, 6, 1), datetime.date(2012, 6, 2)]
        >>> list(date_range(datetime(2012, 6, 2, 23, 59, 59, 999999), datetime(2012, 6, 1, 23, 59, 59, 999999)))
        [datetime.date(2012, 6, 2), datetime.date(2012, 6, 1)]
        >>> list(date_range(datetime(2012, 6, 1, 0, 0), datetime(2012, 6, 1, 23, 59, 59, 999999)))
        [datetime.date(2012, 6, 1)]
    '''

    increasing = start <= end
    day = datetime_to_date(start)

    if lei_convention:
        if increasing:
            if isinstance(end, datetime):
                end = end - one_microsecond
            else:
                end = end - one_day
        else:
            if isinstance(end, datetime):
                end = end + one_microsecond
            else:
                end = end + one_day

    end = datetime_to_date(end)

    yield day
    if increasing:
        while day < end:
            day = day + one_day
            yield day
    else:
        while day > end:
            day = day - one_day
            yield day


class elapsed_time():
    ''' Context manager to compute elapsed time.

        >>> ms = 200
        >>> with elapsed_time() as et:
        ...     time.sleep(float(ms)/1000)
        >>> print(et)
        200 ms
        >>> delta = et.timedelta()
        >>> lower_limit = timedelta(milliseconds=ms)
        >>> upper_limit = timedelta(milliseconds=ms+1)
        >>> assert delta > lower_limit
        >>> assert delta <= upper_limit
    '''

    def __init__(self):
        self.start = now()
        self.end = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.end = now()

    def __str__(self):
        return timedelta_to_human_readable(self.timedelta())

    def timedelta(self):
        ''' Elapsed time as timedelta type.

            If still in block, then elapsed time so far. '''

        if self.end:
            result = self.end - self.start
        else:
            result = now() - self.start
        return result

class log_elapsed_time():
    ''' Context manager to log elapsed time.

        >>> ms = 200
        >>> with log_elapsed_time(log, 'test sleep'):
        ...     time.sleep(float(ms)/1000)
    '''

    def __init__(self, log, msg=None):
        self.start = now()
        # verify 'log' is a log
        if not hasattr(log, 'debug'):
            raise ValueError(f"'log' must be a log, not {type(log)}")
        self.log = log
        self.msg = msg

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        elapsed = now() - self.start
        if self.msg:
            self.log.debug(f'{self.msg} elapsed time {elapsed}')
        else:
            self.log.debug(f'elapsed time {elapsed}')

def date_string_to_date(date_string):
    ''' Convert a string representation of a date into a python date.

        >>> date_string_to_date('2015-04-25')
        datetime.date(2015, 4, 25)
        >>> date_string_to_date('14-01-2015')
        datetime.date(2015, 1, 14)
        >>> date_string_to_date('test')
    '''

    Date_Format1 = r'(\d{4})-(\d{2})-(\d{2})'
    Date_Format2 = r'(\d{2})-(\d{2})-(\d{4})' # the European alternative

    d = None
    m = re.match(Date_Format1, date_string)
    if m:
        d = date(int(m.group(1)), int(m.group(2)), int(m.group(3)))
    else:
        m = re.match(Date_Format2, date_string)
        if m:
            d = date(int(m.group(3)), int(m.group(2)), int(m.group(1)))

    return d


# be careful about using the following values
# they are calculated when this module is imported
today = now()
tomorrow = now() + one_day
yesterday = now() - one_day
one_month_ago = one_month_before(today)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Parse an apache compatible access log line.

    Copyright 2012-2020 DeNova
    Last modified: 2020-10-20
'''

import re
from datetime import datetime, timezone

from denova.net.browser import get_agent_info
from denova.python.dict import Simple
from denova.python.utils import is_string

try:
    from denova.python.log import get_log
    log = get_log()
except:   # 'bare except' because it catches more than "except Exception"
    def log(message):
        print(message)
debug = True
def note(*args):
    print(args)
    log(*args)


class LogLine(Simple):
    ''' Parse a web log line.

    >>> access = LogLine(b'200.217.153.246 - - [10/Dec/2019:03:36:17 +0000] "GET /trader/trial/ HTTP/1.0" 200 85755 "https://denova.com/" "Mozilla/5.0 (Wayland; Linux x86_64; rv:58.0) Gecko/20100101 Firefox/58.0"')

    >>> access.ip_address
    '200.217.153.246'
    >>> access.domain
    ''
    >>> access.user
    ''
    >>> access.timestamp
    datetime.datetime(2019, 12, 10, 3, 36, 17, tzinfo=datetime.timezone.utc)
    >>> access.method
    'GET'
    >>> access.url
    '/trader/trial/'
    >>> access.query_string
    ''
    >>> access.protocol
    'HTTP/1.0'
    >>> access.http_status_code
    200
    >>> access.bytes
    85755
    >>> access.referer
    'https://denova.com/'
    >>> access.agent
    'Mozilla/5.0 (Wayland; Linux x86_64; rv:58.0) Gecko/20100101 Firefox/58.0'


    >>> access = LogLine(b'172.105.23.36 - - [10/Dec/2019:06:01:22 +0000] "GET / HTTP/1.1" 400 173 "-" "-"')
    >>> access.ip_address
    '172.105.23.36'
    >>> access.domain
    ''
    >>> access.user
    ''
    >>> access.timestamp
    datetime.datetime(2019, 12, 10, 6, 1, 22, tzinfo=datetime.timezone.utc)
    >>> access.method
    'GET'
    >>> access.url
    '/'
    >>> access.query_string
    ''
    >>> access.protocol
    'HTTP/1.1'
    >>> access.http_status_code
    400
    >>> access.bytes
    173
    >>> access.referer
    ''
    >>> access.agent
    ''

    >>> access = LogLine(b'184.156.108.45 - - [10/Dec/2019:19:25:50 +0000] "GET /rescue/downloads/denova-rescue_0.9-1_all.deb HTTP/1.1" 200 7186896 "https://denova.com/rescue/downloads/" "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0"')
    >>> access.ip_address
    '184.156.108.45'
    >>> access.domain
    ''
    >>> access.user
    ''
    >>> access.timestamp
    datetime.datetime(2019, 12, 10, 19, 25, 50, tzinfo=datetime.timezone.utc)
    >>> access.method
    'GET'
    >>> access.url
    '/rescue/downloads/denova-rescue_0.9-1_all.deb'
    >>> access.query_string
    ''
    >>> access.protocol
    'HTTP/1.1'
    >>> access.http_status_code
    200
    >>> access.bytes
    7186896
    >>> access.referer
    'https://denova.com/rescue/downloads/'
    >>> access.agent
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0'

    >>> line = b'216.244.66.196 - - [10/Dec/2019:06:21:47 +0000] "GET /trader/opportunities/ HTTP/1.1" 200 86165 "-" "BusinessBot: Nathan@lead-caddy.com"'
    >>> access = LogLine(line)
    >>> access.agent
    'BusinessBot: Nathan@lead-caddy.com'
    >>> access.browser_name
    ''
    >>> access.other
    'BusinessBot: Nathan@lead-caddy.com'
    '''

    """ IMPORTANT: Be careful not to start or end any lines with triple quotes,
        unless the quoted string is a comment.
        Our code that strips code comments for distribution may break.
        For example, the following regular expression *won't work* unless the r'''
        is moved to the next line, after leading spaces and the ''', line
        can't start the line:

        Bad:
            Line_Format = re.compile(r'''
                                     (?P<ip_address>.*?),?\s"
                                     ''',
                                     re.VERBOSE)
        Good (at least it works with the comment stripping code):
            Line_Format = re.compile(
                                     r'''
                                     (?P<ip_address>.*?),?\s" ''',
                                     re.VERBOSE)
    """
    Line_Format = re.compile(
        r'''(?P<ip_address>.*?),?\s

        (
            (-) # missing domain
            |
            (
                (?P<domain>.+?)
            )
        )\s

        (
            (-) # missing user
            |
            (
                (?P<user>.+?)
            )
        )\s

        \[
            (?P<timestamp>.*?)
        \]\s

        "
            (
                (-) # missing method/url/protocol
                |
                (
                    (?P<method>.+?)\s
                    (?P<url>.+?)
                    (\?(?P<query_string>.+?))?\s
                    (?P<protocol>.+?)
                )
            )
        "
        \s

        (?P<http_status_code>\d+)\s
        (?P<bytes>.+?)\s

        "
            (
                (-) # missing referer
                |
                (
                    (?P<referer>.+?)
                )
            )
        "\s

        "
            (
                (-) # missing agent
                |
                (
                    (?P<agent>.+?)
                )
            )
        " ''',
        re.VERBOSE)
    # old: Line_Format = re.compile(r'''(?P<ip_address>.*?),? (?P<domain>.*?) (?P<user>.*?) \[(?P<timestamp>.*?)\] "(?P<method>.*?) (?P<url>.*?)(\?(?P<query_string>.*?))? (?P<protocol>.*?)" (?P<http_status_code>\d*) (?P<bytes>.*?) "(?P<referer>.*?)" "(?P<agent>.*?)"''')
    # testing 2: Line_Format = re.compile(r'''(?P<ip_address>.*?),? (?P<domain>.*?) (?P<user>.*?) \[(?P<timestamp>.*?)\] "((?P<method>.*?) (?P<url>.*?)(\?(?P<query_string>.*?))? (?P<protocol>.*?))?" (?P<http_status_code>\d*) (?P<bytes>.*?) "(?P<referer>.*?)" "(?P<agent>.*?)"''')

    Date_Format = re.compile(r'''(?P<day>\d*)/(?P<month>.*)/(?P<year>\d*):(?P<hour>\d*):(?P<min>\d*):(?P<sec>\d*) *(?P<gmt_offset>.*)''')
    Page_Not_Found_Format = re.compile(r'''(?P<ip_address>.*?),? (?P<domain>.*?) (?P<user>.*?) \[(?P<timestamp>.*?)\] ".*?" 400 \d{1,3} "-" "-"''')
    #Ip_Address_Format = re.compile(r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}")

    NOT_APACHE_FORMAT = 'Not in apache log format'

    def __init__(self, line):
        ''' Line is a line from an apache compatible web log. '''

        self.line = line

        m = LogLine.Line_Format.search(str(line))
        if m:
            self.ip_address = LogLine.get_string(m.group('ip_address'))
            if not is_string(self.ip_address):
                self.ip_address = self.ip_address.decode()
            self.domain = LogLine.get_string(m.group('domain'))
            self.method = LogLine.get_string(m.group('method'))
            if self.method:
                self.method = self.method.upper()

            self.timestamp = LogLine.get_timestamp(m.group('timestamp'))

            self.referer = LogLine.get_string(m.group('referer'))
            if self.referer and self.referer == '-':
                self.referer = ''

            self.url = LogLine.get_string(m.group('url'))
            self.query_string = LogLine.get_string(m.group('query_string'))
            self.http_status_code = LogLine.get_int(m.group('http_status_code'))
            self.protocol = LogLine.get_string(m.group('protocol'))
            self.bytes = LogLine.get_int(m.group('bytes'))

            self.user = LogLine.get_string(m.group('user'))
            if self.user and self.user == '-':
                self.user = ''

            self.agent = LogLine.get_string(m.group('agent'))
            self.browser_name, self.browser_version, self.other = LogLine.get_browser_info(self.agent)

        else:
            message = f'{LogLine.NOT_APACHE_FORMAT}: {line}'

            if not is_string(line):
                line = line.decode()
            m = LogLine.Page_Not_Found_Format.search(line)
            if not m:
                log(message)
            # but still raise the error so we don't process the entry
            raise ValueError(message)

    @staticmethod
    def get_string(item):
        '''Get a stripped string from the item.

           If the item is None, then return an empty string.
       '''

        string = item
        if item is None:
            string = ''
        else:
            # some strings have what looks like embedded byte code
            string = string.strip().lstrip("b'")

        return string

    @staticmethod
    def get_int(item):
        if item:
            try:
                bytes = int(item)
            except:   # 'bare except' because it catches more than "except Exception"
                bytes = None
        else:
            bytes = None

        return bytes

    @staticmethod
    def get_timestamp(item):
        '''Parse the date/time from the string.'''

        # import delayed to avoid recusive import
        import denova.python.times

        if item:

            m = LogLine.Date_Format.search(item)
            if m:
                hour = int(m.group('hour'))
                minutes = int(m.group('min'))
                seconds = int(m.group('sec'))
                day = int(m.group('day'))
                month = denova.python.times.MONTH_MAP[m.group('month')]
                year = int(m.group('year'))

                date_time = datetime(year, month, day, hour, minutes, seconds, tzinfo=timezone.utc)

            else:
                date_time = None

        else:
            date_time = None

        return date_time

    @staticmethod
    def get_browser_info(item):
        '''Parse the browser info from the string.

            >>> LogLine.get_browser_info('Mozilla/5.0 (compatible; Yahoo! Slurp/3.0; http://help.yahoo.com/help/us/ysearch/slurp)')
            ('Mozilla', '5.0', '(compatible; Yahoo! Slurp/3.0; http://help.yahoo.com/help/us/ysearch/slurp)')
            >>> LogLine.get_browser_info('BusinessBot: Nathan@lead-caddy.com')
            ('', '', 'BusinessBot: Nathan@lead-caddy.com')
        '''

        browser_name, browser_version, other = get_agent_info(item)

        return (browser_name, browser_version, other)

    @staticmethod
    def full_url(access):
        ''' Return the full url with query string.'''

        url = access.url
        if access.query_string:
            url += f'?{access.query_string}'

        return url


if __name__ == "__main__":

    import doctest
    doctest.testmod()




'''
    Build virtualenv for one of our websites.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-07

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

from abc import ABCMeta, abstractmethod
import os
import sys
from subprocess import CalledProcessError
from traceback import format_exc

from denova.os.command import run
from denova.python.log import get_log
from denova.python.ve import venv


class BuildVenv(metaclass=ABCMeta):
    ''' Build the virtualenv for a site.

        Each site should inherit from this abstract class with details about their environment.
    '''

    VIRTUAL_SUBDIR = 'virtualenv3'

    PROJECTS_DIR = os.path.realpath(os.path.join(os.path.abspath(os.path.dirname(__file__).replace('\\','/')), '..', '..'))

    def __init__(self):
        self.log = get_log()

    def build(self):

        os.chdir(self.parent_dirname())
        self.init_virtualenv()

        if self.virtualenv_dir_exists():
            self.log(f'building virtual environment in {self.parent_dirname()}')
            os.chdir(self.virtualenv_dir())

            # activate the virtualenv
            with venv(dirname=self.virtualenv_dir()):

                # set up a link to the python lib for simplier config
                dirname = None
                entries = os.scandir(os.path.join(self.virtualenv_dir(), 'lib'))
                for entry in entries:
                    if entry.name.startswith('python3'):
                        dirname = entry.name
                        break
                if dirname is None:
                    dirname = 'python3.7'

                os.chdir('lib')
                run('ln', '-s', dirname, 'python')

                self.report('   installing requirements')
                with open(self.get_requirements()) as f:
                    for line in f.readlines():
                        if line.strip():
                            app = line.strip()
                            self.report(f'     {app}')
                            try:
                                run('pip3', 'install', app)
                            except CalledProcessError as cpe:
                                self.log(format_exc())
                                if cpe.stdout: self.log(f'stdout: {cpe.stdout}')
                                if cpe.stderr: self.log(f'stderr: {cpe.stderr}')
                                sys.exit(
                                  f'{cpe.stderr}. For more details see {self.log.pathname}')

            self.log('   linking packages')
            self.link_packages(os.path.join(self.virtualenv_dir(), 'lib', 'python', 'site-packages'))

            self.finish_build()
            self.log('   virtual environment built')
        else:
            self.log(f'!!!Error: Unable to create {self.virtualenv_dir()}')

    @abstractmethod
    def parent_dirname(self):
        ''' Directory where virtualenv will be created. '''

    @abstractmethod
    def get_requirements(self):
        ''' Return the full path to the virtualenv requirements. '''

    @abstractmethod
    def link_packages(self, site_packages_dir):
        ''' Link packages to the site-packages directory. '''

    @abstractmethod
    def virtualenv_dir(self):
        ''' Returns the full path to the virtualenv directory. '''

    def virtual_subdir(self):

        return self.VIRTUAL_SUBDIR

    def init_virtualenv(self):
        '''
            Initialize the virtualenv.

            Overwrite this function if you want
            special switches used when running virtualenv.
        '''

        if os.path.exists(self.virtual_subdir()):
            run('rm', '-fr', self.virtual_subdir())
        if os.path.exists('/usr/bin/python3.7'):
            python_pgm = 'python3.7'
        elif os.path.exists('/usr/bin/python3.5'):
            python_pgm = 'python3.5'
        else:
            entries = os.scandir('/usr/bin')
            for entry in entries:
                if entry.name.startswith('python3'):
                    python_pgm = entry.name
                    break
        run('virtualenv', '-p', f'/usr/bin/{python_pgm}', self.virtual_subdir())

    def virtualenv_dir_exists(self):
        ''' Return True if the virtualenv directory does exist. '''

        return os.path.exists(self.virtualenv_dir())

    def finish_build(self):
        ''' Overwrite if there are any final steps necessary to create the build.'''
        # do not remove the follow pass; when we strip comments, we need the pass
        pass

    def report(self, message):
        ''' Show and log the message. '''

        print(message)
        self.log(message)



'''
    Extra html functions.
    html_addons is named to avoid conflict with python's html pacakge.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-20

    Requires BeautifulSoup, html5lib, and lxml for proper pretty printing of HTML.

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import re

from denova.python.format import to_bytes, to_string
from denova.python.log import get_log

log = get_log()

def extract_text(html):
    ''' Extract plain text from html.

        Prefers BeautifulSoup 4, but falls back to ad hoc extraction.
    '''

    try:
        from bs4 import BeautifulSoup

    except ImportError:
        # ad hoc
        #log.debug('html: {}'.format(repr(html)))
        text = re.sub(r'<.*?>', ' ', html.strip())
        #log.debug('text after re: {}'.format(text))
        text = text.replace('  ', ' ')
        # some html tags are still present; overlapping regex matches?
        # throw away anything after the first '<'
        text, _, _ = text.partition('<')
        #log.debug('final text: {}'.format(text))

    else:
        # BeautifulSoup
        # from http://stackoverflow.com/questions/1936466/beautifulsoup-grab-visible-webpage-text
        soup = BeautifulSoup(html, features='lxml')
        texts = soup.findAll(text=True)

        def visible(element):
            if element.parent.name in ['style', 'script', '[document]', 'head', 'title']:
                return False
            elif re.match('<!--.*-->', str(element)):
                return False
            return True

        visible_texts = list(filter(visible, texts))

        text = '\n'.join(visible_texts)

    return text

def get_links(htmlpath, exclude=None):
    ''' Get links from an html file.

        Not well tested. See reinhardt.feeds for examples of more reliable parsing.

        Returns a list. Each item is a list of [PATH, URL, SUMMARY].

        'htmlpath' is path of html file.

        'exclude' is string in href to exclude, without top level domain.
        Example: To exclude links to google, use "exclude='google'".

        Very ad hoc.
    '''

    # fallable importdelayed until needed
    try:
        from pyquery.pyquery import PyQuery

    except ModuleNotFoundError:
        raise Exception('pyquery not installed')

    else:

        results = []

        with open(htmlpath) as infile:

            html = PyQuery(to_bytes(infile.read()))
            anchor_tags = html.items('a')
            # log.debug('{} links: {}'.format(len(list(anchor_tags)), htmlpath)) # DEBUG
            for item in anchor_tags:
                href = item.attr('href')
                if href and href.startswith('http'):
                    if exclude and (exclude not in href):
                        results.append([htmlpath, href, item.text().strip()])
                        # log.debug('\t{}'.format(href)) # DEBUG

        return results

def expose_hidden_tags(html):
    ''' Make spoofed tags explicit. '''

    # this may not work with some cases. Example:
    #     <description>&lt; . . . &gt;</description>
    # a possible workaround is to re-encode these cases of '<' and '>' after firewalling

    if isinstance(html, str):
        start_tag = '<'
        end_tag = '>'
        hidden_start_tag = '&lt;'
        hidden_end_tag = '&gt;'
    elif isinstance(html, bytes):
        start_tag = b'<'
        end_tag = b'>'
        hidden_start_tag = b'&lt;'
        hidden_end_tag = b'&gt;'
    else:
        raise ValueError('html must be a string or bytes')

    if hidden_start_tag in html:
        log.warning(f'exposing {str(hidden_start_tag)} in html')
        html = html.replace(hidden_start_tag, start_tag)
    if hidden_end_tag in html:
        log.warning(f'exposing {str(hidden_end_tag)} in html')
        html = html.replace(hidden_end_tag, end_tag)

    return html

def clean_xml(xml):
    ''' Return (more) valid xml. Html is xml, so clean_xml() also cleans html.

        BeautifulSoup has a rep for accepting bad xml.
        Then it can write good xml.
    '''

    try:
        from bs4 import BeautifulSoup
    except ImportError:
        log('BeautifulSoup not installed so xml cannot be cleaned.')
        raise
    else:
        try:
            soup = BeautifulSoup(xml, features='html5lib')
        except:  # 'bare except' because it catches more than "except Exception"
            # "features='lxml'" is to silence a very noisy and useless
            # error message from BeautifulSoup
            soup = BeautifulSoup(xml, features='lxml')

        cleaned_xml = soup.prettify()

    return cleaned_xml

def clean_html(html):
    ''' Return (more) valid html.

        Convenience function for clean_xml(html).
    '''

    return clean_xml(html)

def remove_doctype(html):
    ''' Remove all occurrences of DOCTYPE from html/xml.

        Some invalid html/xml has more than one DOCTYPE.
        A simple approach is to remove them both.

        This appears to be unnecessary if we call clean_xml().
        BeautifulSoup removes the extra DOCTYPE.

        This function actually removes all DOCTYPE headers, which
        appears to be overkill.
    '''

    DOCTYPE = r'\<!\s*DOCTYPE.*?\>'
    return re.sub(DOCTYPE, '', html, flags=re.IGNORECASE)

def find_tags(elements, tags, matches=None):
    ''' Find all matching tags in xmltodict elements. '''

    DEBUG = True

    def debug(msg):
        if DEBUG:
            log(f'find_tags: {msg}')

    # if single tag, make it a sequence
    if isinstance(tags, str):
        tags = [tags]

    # lower case tags
    tags = [tag.lower() for tag in tags]

    debug(f'find tags matching {tags}') # DEBUG

    # if this is the top of the find_tags() recursion chain
    if matches is None:
        matches = []

    for key in elements:

        debug(f'subkey: {key}') # DEBUG
        value = elements[key]

        if key in tags:
            debug(f'subkey matches tag: {key}') # DEBUG
            if isinstance(value, list):
                for v in value:
                    matches.append(v)
            else:
                matches.append(value)

        else:
            if isinstance(value, dict):
                debug('recursing') # DEBUG
                matches = matches + find_tags(value, tags, matches)
                debug('back from recursion') # DEBUG

                # xmltodict.unparse(value, pretty=True)
                keys = []
                for subkey in value:
                    keys.append(subkey)
                if keys:
                    debug(f'subkeys for "{key}":\n{keys}')

            elif isinstance(value, str):
                debug(f'subkey value: "{value}"')

            elif isinstance(value, list):
                for item in value:
                    debug('recursing') # DEBUG
                    matches = matches + find_tags(value, tags, matches)
                    debug('back from recursion') # DEBUG

            else:
                debug(f'subkey value type: {type(value)}')
                debug(f'subkey value: {repr(value)}')

    return matches

def count_tags(elements, tag):
    ''' Count matching tags in elements. '''

    return len(find_tags(elements, tag))

def pretty_element(element):
    ''' Intended to return pretty-printed html element.
        But because we're using pyquery, just returns html.
    '''

    return element.html()
    # return to_string(lxml.etree.tostring(element, pretty_print=True)).strip()

def decode_html_file(path):
    ''' Decode an html file using the correct encoding.

        Any special encoding should be in the http header.
        But the requests lib would have already handled that.
        Sometimes an html/xml file will specify its encoding in
        an xml comment at the top of the file. That's what we
        are trying to handle here.
    '''

    log(f'UnicodeDecodeError: {path}')

    log(f'reread as bytes: {path}')
    # and try to decode
    with open(path, 'rb') as infile:
        content = infile.read()

    log(f'look for xml encoding: {path}')
    encoding = get_xml_encoding(content)
    if encoding:
        try:
            content = content.decode(encoding)
        except Exception:
            # the declared encoding may be wrong
            # try the default utf-8
            # this i
            content = content.decode()

        # lxml won't accept bytes with an encoding declaration
        log(f'strip any xml declaration: {path}')
        if content.strip().startswith('<?'):
            _, separator, suffix = content.partition('?>')
            if separator:
                content = suffix

    else:
        log.warning(f'UnicodeDecodeError, but no encoding specified: {path}')
        # to_string() tries common encodings
        content = to_string(content)
        log('successfully decoded content')

    return content

def get_xml_encoding(content):
    '''
        Get character encoding from xml header.
    '''

    encoding = None
    if content.strip().startswith(b'<?'):
        prefix, separator, _ = content.partition(b'?>')
        if separator:
            xml_declaration = prefix + separator
            log.debug(f'xml_declaration: {xml_declaration}')

            match = re.search(rb'encoding\s*=\s*[\"\'](.+?)[\"\']',
                              xml_declaration,
                              flags=re.IGNORECASE)
            if match:
                encoding = match.group(1)
                # log.debug(f'{self.name} encoding: {encoding}')

    return encoding

def get_title(html):
    ''' Get title from html.

        Return title, or None if no title found.

        >>> html = """
        ...     <html lang="en">
        ...         <head>
        ...              <title>
        ...                  Test title
        ...              </title>
        ...
        ...              <meta charset="utf-8">
        ...              </meta>
        ...         </head>
        ...
        ...         <body>
        ...             just a test
        ...         </body>
        ...
        ...     </html>
        ... """

        >>> print(get_title(html))
        Test title
    '''

    PATTERN = re.compile(r'''<\s*title\s*.*?>
                                 (.*?)
                             <\s*/title\s*>
                         ''',
                         re.VERBOSE | re.DOTALL | re.IGNORECASE)
    match = PATTERN.search(html)
    if match:
        title = match.group(1).strip()
    else:
        title = None

    return title

def strip_whitespace_in_html(value, mintext=False):
    ''' Returns the given HTML with minimum white space.

        To also change white space sequences in text between tags
        to a single space, set mintext to True. This is dangerous
        with embedded javascript, <pre>, etc.

        >>> strip_whitespace_in_html(' <a>   <b> test test2 </b></a>')
        '<a><b> test test2 </b></a>'
    '''

    value = value.strip()
    # no spaces around tags
    value = re.sub(r'>[\s\n\t]+<', '><', value)
    if mintext:
        # other space sequences replaced by single space
        value = re.sub(r'\s+', ' ', value)

    return value

def strip_cdata(s):
    ''' Strip CDATA markup. '''

    # Example:
    #     <![CDATA[Text we want to extract]]>
    return re.sub(r'<\s*!\s*\[\s*CDATA\s*\[(.*?)\s*]\s*]>',
                  r'\1',
                  s,
                  flags=re.IGNORECASE)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python3
'''
    File system.

    Copyright 2008-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os
import os.path
import shutil
import stat
import tempfile
import threading
from contextlib import contextmanager
from datetime import datetime
from stat import S_ISDIR
from subprocess import CalledProcessError

from denova.os.command import run
from denova.python.log import get_log
from denova.python.utils import is_string

log = get_log()

empty_dir = None

DEFAULT_PERMISSIONS_DIR_OCTAL = 0o755
DEFAULT_PERMISSIONS_FILE_OCTAL = 0o640

# use strings with chmod command, run('chmod'), and denova.os.fs.chmod()
DEFAULT_PERMISSIONS_DIR = 'u=rwX,g=rX,o=rX'
DEFAULT_PERMISSIONS_FILE = 'u=rw,g=r,o='
''' This produces an unreadable int and ignores the difference in
    chmod between 'x' and 'X'.
try:
    DEFAULT_PERMISSIONS_DIR = (
        stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
        stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP |
        stat.S_IROTH | stat.S_IXOTH)
    assert DEFAULT_PERMISSIONS_DIR == DEFAULT_PERMISSIONS_DIR_OCTAL
    DEFAULT_PERMISSIONS_FILE = (
        stat.S_IRUSR | stat.S_IWUSR |
        stat.S_IRGRP | stat.S_IWGRP |
        stat.S_IROTH | stat.S_IWOTH)
    assert DEFAULT_PERMISSIONS_FILE == DEFAULT_PERMISSIONS_FILE_OCTAL
except:   # 'bare except' because it catches more than "except Exception"
    DEFAULT_PERMISSIONS_DIR = DEFAULT_PERMISSIONS_DIR_OCTAL
    DEFAULT_PERMISSIONS_FILE = DEFAULT_PERMISSIONS_FILE_OCTAL
'''

class FileSystemException(Exception):
    pass

@contextmanager
def cd(dirname):
    ''' Context manager to change dir temporarily.

        >>> startdir = '/tmp'
        >>> os.chdir(startdir)
        >>> assert os.getcwd() == startdir
        >>> tempdir = tempfile.mkdtemp()
        >>> with cd(tempdir):
        ...     assert os.getcwd() == tempdir
        ...     assert os.getcwd() != startdir
        >>> assert os.getcwd() == startdir

    '''

    olddir = os.getcwd()
    os.chdir(dirname)
    try:
        yield
    finally:
        try:
            os.chdir(olddir)
        except Exception as ex:
            """ Example:
                  1. root's cwd is /root
                  2. root invokes sudo to run a program that calls cd()
                  3. The user doesn't have permission to cd back to /root
                Workaround is to cd to a safe dir before sudo command.
            """
            log.error(f'could not cd from {dirname} to {olddir}')
            log.error(ex)
            raise

def chmod(mode, path, recursive=False):
    ''' Change permissions of path.

        mode is a string or octal integer for the chmod command.
        Examples::
            'o+rw,g+rw'
            0660
    '''

    # arg order used to be chmod(path, mode, ...), so check types
    # delete this assert if no assertion errors 2015-01-01
    # after we remove this assert, mode can be a string
    assert is_string(path)

    if isinstance(mode, int):
        # chmod wants an octal int, not decimal
        # so if it's an int we convert to an octal string
        oct_mode = oct(mode)
        if 'o' in oct_mode:
            i = oct_mode.find('o')
            oct_mode = oct_mode[i+1:]

        if len(oct_mode) > 4 and not oct_mode.startswith('0'):
            mode = '0' + oct_mode
        else:
            mode = oct_mode

    try:
        if recursive:
            run(*['chmod', '--recursive', mode, path])
        else:
            run(*['chmod', mode, path])
    except CalledProcessError as cpe:
        log.error(f'unable to chmod: path={path}, mode={mode}')
        log.error(cpe)
        raise

def chown(owner, path, recursive=False):
    ''' Change owner of path.

        'owner' can be the user name, uid as an int, or uid as a string.

        Log and reraise any exception.
    '''

    try:
        if recursive:
            run(*['chown', '--recursive', owner, path])
        else:
            run(*['chown', owner, path])
        #log.debug('chown(owner={}, path=={})'.format(owner, path))
    except CalledProcessError as cpe:
        # import delayed to avoid infinite recursion
        from denova.os.user import whoami

        log.error(f'unable to chown: user={whoami()}, owner={owner}, path={path}')
        log.error(cpe)
        raise

    # verify. after we have higher confidence, move this into doctests
    if type(owner) is str and ':' in owner:
        owner, group = owner.split(':')
    else:
        group = None
    try:
        uid = int(owner)
    except ValueError:
        # import delayed to avoid infinite recursion
        from denova.os.user import getuid

        uid = getuid(owner)
    assert getuid(path) == uid, f'uid set to {uid} but is {getuid(path)}'
    if group is not None:
        try:
            gid = int(group)
        except ValueError:
            # import delayed to avoid infinite recursion
            from denova.os.user import getuid

            gid = getgid(group)
        assert getgid(path) == gid, f'gid set to {gid} but is {getgid(path)}'

def chgrp(group, path, recursive=False):
    ''' Change group of path.

        'group' can be the group name, gid as an int, or gid as a string.

        Log and reraise any exception.
    '''

    try:
        if recursive:
            run(*['chgrp', '--recursive', group, path])
        else:
            run(*['chgrp', group, path])
        #log.debug('chgrp(group={}, path=={})'.format(group, path))
    except CalledProcessError as cpe:
        # import delayed to avoid infinite recursion
        from denova.os.user import whoami

        log.error(f'unable to chgrp: user={whoami()}, group={group}, path={path}')
        log.error(cpe)
        raise

    # verify. after we have higher confidence, move this into doctests
    if type(group) is str and ':' in group:
        owner, group = group.split(':')
    else:
        owner = None
    if owner is not None:
        try:
            uid = int(owner)
        except ValueError:
            # import delayed to avoid infinite recursion
            from denova.os.user import getuid

            uid = getuid(owner)
        assert getuid(path) == uid, f'uid set to {uid} but is {getuid(path)}'
    try:
        gid = int(group)
    except ValueError:
        # import delayed to avoid infinite recursion
        from denova.os.user import getuid

        gid = getgid(group)
    assert getgid(path) == gid, f'gid set to {gid} but is {getgid(path)}'

def getmode(path):
    ''' Return permissions (mode) of a path.

        >>> oct(getmode('/var/local/projects/denova/os/fs.py'))
        '0o644'
    '''

    return stat.S_IMODE(os.lstat(path)[stat.ST_MODE])

def getuid(path):
    ''' Return uid of a path. '''

    return os.stat(path).st_uid

def getgid(path):
    ''' Return gid of a path. '''

    return os.stat(path).st_gid

def set_attributes(path, owner=None, group=None, perms=None, recursive=False):
    ''' Set ownership and permissions. '''

    if owner is not None:
        chown(owner, path, recursive)
    if group is not None:
        chgrp(group, path, recursive)
    if perms is not None:
        chmod(perms, path, recursive)

def makedir(dirname, owner=None, group=None, perms=None):
    ''' Make dir with default ownership and permissions.

        Makes parent dirs if needed. If dir already exists, ownership
        and permissions are not changed.

        Use this instead of os.mldir() and os.makedirs().
        Those methods mask out the current umask value
    '''

    if perms is None:
        perms = DEFAULT_PERMISSIONS_DIR

    lock = threading.Lock()
    lock.acquire(1) # 1 means blocking

    try:
        if not os.path.exists(dirname):

            # makedirs(dirname, mode) uses umask, so we also call chmod() explicitly
            #log.debug('makedir(dirname={}, owner={}, group={}, perms={})'.format(
            #    dirname, owner, group, oct(perms))) #DEBUG
            try:
                os.makedirs(dirname)
            except OSError:
                log.error(why_file_permission_denied(dirname, perms))
                raise
            else:
                set_attributes(dirname, owner, group, perms, recursive=True)

    finally:
        lock.release()

    assert os.path.isdir(dirname), f'could not make dir: {dirname}'

@contextmanager
def temp_mount(*args, **kwargs):
    ''' Context manager to mount/unmount a filesystem.

        Example::
            >>> from denova.os.user import whoami
            >>> if whoami == 'root':
            ...     mountpoint = tempfile.mkdtemp()
            ...     with temp_mount('/tmp', mountpoint):
            ...         mounted(mountpoint)
            ... else: print(True)
            True
            >>> if whoami == 'root':
            ...     mounted(mountpoint)
            ... else: print(False)
            False
            >>> if whoami == 'root':
            ...     os.rmdir(mountpoint)

        Mounts/unmounts only if was unmounted.
        Leaves the mount point in the original state, mounted or unmounted,
        unless the context block changes the mount state.
    '''

    ''' Needs more tests, especially for --bind/move/make-xxx with umount(args[-1]) '''

    # the last arg is reliably the mount point
    # even with --bind, --move, etc.
    mountpoint = os.path.abspath(args[-1])
    if mounted(mountpoint):
        yield

    else:
        # import delayed to avoid infinite recursion
        from denova.os.user import sudo

        with sudo():
            # last arg is mount point
            log.debug(f'temp mount {mountpoint}')
            mount(*args, **kwargs)

        try:
            yield

        finally:
            # import delayed to avoid infinite recursion
            from denova.os.user import sudo

            with sudo():
                log.debug(f'ummount temp mount {mountpoint}')
                unmount(mountpoint)

def mounts():
    ''' Return filesystem mounts as list.

        List elements are (device, mountpoint, vfstype, options).
        All elements are strings.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     m = mounts()
        ...     for knownmount in ['/', '/sys', '/proc', '/dev']:
        ...         assert any(knownmount == mountpoint for (device, mountpoint, vfstype, options) in m)
    '''

    # import delayed to avoid infinite recursion
    from denova.os.user import sudo


    results = []
    with sudo():
        with open('/etc/mtab') as mtab:
            for line in mtab:
                device, mountpoint, vfstype, options, dump, fsck = tuple(line.split())
                results.append((device, mountpoint, vfstype, options))

    return results

def mounted_devices():
    ''' Return list of mounted devices. '''

    return [device
        for device, mountpoint, vfstype, options in mounts()]

def unmounted_block_devices():
    ''' Return list of unmounted devices that may have filesystems,
        i.e. unmounted block devices. '''

    unmounted = []
    mounted = mounted_devices()
    for device in devices():
        # /dev/sda may appear to not be mounted, but /dev/sda1 is
        dev_mounted = False
        for mounted_device in mounted:
            if mounted_device.startswith(device):
                dev_mounted = True
        if not dev_mounted:
            unmounted.append(device)

    return unmounted

def mountpoints():
    ''' Return list of filesystem mountpoints.

        >>> from denova.os.user import whoami
        >>> if whoami() == 'root':
        ...    for knownmount in ['/', '/sys', '/proc', '/dev']:
        ...        assert knownmount in mountpoints()
    '''

    return [mountpoint
        for device, mountpoint, vfstype, options in mounts()]

def mount(*args, **kwargs):
    ''' Convenience function for mount. '''

    try:
        all_args = ['mount']
        for arg in args:
            all_args.append(arg)
        run(*all_args, **kwargs)
    except Exception as exc:
        log.debug(exc)
        raise

def unmount_all(dirname, force=False):
    ''' Unmount all points in the named dir.

        If force=True, unmount even if mounts are in use.
    '''

    log.debug(f'unmount all in {dirname}')
    # make sure we have a trailing slash
    dir_prefix = os.path.join(dirname, '')

    # umount last mounted first, dirname last
    try:
        for device, mountpoint, vfstype, options in reversed(mounts()):
            if mountpoint.startswith(dir_prefix):
                unmount(mountpoint, force=force)
    except FileSystemException:
        # still try umount at the top level dir
        pass

    unmount(dirname)

def unmount(mountpoint, force=False):
    ''' Unmount mountpoint.

        If force=True, unmount even if mounts are in use.'''

    if mounted(mountpoint):
        log.debug(f'unmount {mountpoint}')

        try:
            run(*['umount', mountpoint])

        except Exception as exc:
            log.debug(f'error in first attempt to unmount: {mountpoint}')
            log.debug(exc)

            # if still mounted, try to find out why
            if mounted(mountpoint):

                # import delayed to avoid import recursion
                import denova.python.process

                # find what process has the path open
                msg = f'unmount failed: {mountpoint}'
                programs = denova.python.process.programs_using_file(mountpoint)
                if programs:
                    msg += f', in use by {programs}'
                log.debug(msg)

                # try waiting a couple of seconds
                import time # DEBUG
                time.sleep(2) # DEBUG
                if mounted(mountpoint): # DEBUG
                    log.debug(f'still mounted after delay: {mountpoint}') # DEBUG
                    if force:
                        # 'lazy' unmount
                        # see linux - Umount a busy device - Stack Overflow
                        #     http://stackoverflow.com/questions/7878707/umount-a-busy-device
                        log.warning(f"force=True, so trying 'lazy' unmount: {mountpoint}")
                        run(*['umount', '-l', mountpoint])
                        log.debug('delay after lazy unmount')
                        time.sleep(10)

                        if mounted(mountpoint): # DEBUG
                            log.debug(f'still mounted after forced unmont and delay: {mountpoint}') # DEBUG
                            # see linux - Umount a busy device - Stack Overflow
                            #     http://stackoverflow.com/questions/7878707/umount-a-busy-device
                            log.warning(f"force=True and 'lazy' unmount didn't work, so forcing unmount: {mountpoint}")
                            run(*['umount', '-f', mountpoint])

                else: # DEBUG
                    log.debug(f'NOT mounted after delay: {mountpoint}') # DEBUG

                if mounted(mountpoint):
                    raise FileSystemException(msg)

            else:
                log.debug(f'umount had error but path is not mounted: {mountpoint}')

def mounted(path):
    ''' Return True iff path mounted. '''

    return path.rstrip('/') in mountpoints()

def mounted_on(path):
    ''' Return device mounted on path, or None if none. '''

    mounted_device = None
    for device, mountpoint, vfstype, options in mounts():
        # return first matching device
        if mounted_device is not None:
            if path == mountpoint:
                mounted_device = device

    return mounted_device

def devices():
    ''' Return list of devices that may have filesystems. '''

    # block devices may have filesystems
    raw_output = run(*['lsblk', '--noheadings', '--list', '--paths', '--output=NAME']).stdout.decode()
    devices = raw_output.strip().split('\n')
    return devices

def match_parent_owner(path, mode=None):
    ''' Chown to the parent dir's uid and gid.

        If mode is present, it is passed to chmod(). '''

    try:
        statinfo = os.stat(os.path.dirname(path))
        os.chown(path, statinfo.st_uid, statinfo.st_gid)
    except:   # 'bare except' because it catches more than "except Exception"
        # import delayed to avoid infinite recursion
        from denova.os.user import whoami

        # probably don't have the right to change the file at all,
        # or to change the owner to the specified user
        log.error('unable to chown: user={}, uid={}, gid={}, path={}'.
            format(whoami(), statinfo.st_uid, statinfo.st_gid, path))
        pass
    finally:
        if mode is not None:
            chmod(mode, path)

def get_unique_filename(dirname, prefix, suffix):
    '''
        Get a unique filename.

        >>> dirname = '/tmp'
        >>> prefix = 'test'
        >>> suffix = 'txt'
        >>> filename = get_unique_filename(dirname, prefix, suffix)
        >>> len(filename) > 0
        True
        >>> filename = get_unique_filename(dirname, prefix, suffix)
        >>> len(filename) > 0
        True
        >>> filename = get_unique_filename(dirname, prefix, suffix)
        >>> len(filename) > 0
        True
    '''

    now = datetime.now()
    base_filename = '%s-%d-%02d-%02d-%02d-%02d-%02d' % (
     prefix, now.year, now.month, now.day, now.hour, now.minute, now.second)
    filename = f'{base_filename}.{suffix}'

    if os.path.exists(os.path.join(dirname, filename)):
        i = 1
        filename = f'{base_filename}-{i:02d}.{suffix}'
        while os.path.exists(os.path.join(dirname, filename)):
            i += 1
            filename = f'{base_filename}-{i:02d}.{suffix}'

    return os.path.join(dirname, filename)

def copy(source, dest, symlinks=True, ignore=None, owner=None, group=None, perms=None):
    ''' Copy source to dest dir.

        copy() tries to generally follow the behavior of the
        cp command.

        copy() creates any missing parent dirs of dest.
        File ownership and permissions are also copied.

        If dest is a dir, source is copied into dest.
        An exception is when dest and source have the same basename.
        This is a very common error when the intent is to replace the dest.
        In this case the contents of source are copied to dest.
        This is different from merge() because copy() always deletes the
        dest dir. If you really want a dest named "basename/basename",
        specify it explicitly.

        Otherwise, like cp, source will overwrite any existing dest.

        Unlike shutil.copytree() the default is symlinks=True. So symlinks
        are copied as symlinks. If you want the links replaced by the
        content, set symlinks=False.

        Set ownership and permissions from keywords, or if None copy from
        the corresponding source file.

        >>> testdir = '/tmp/testcopy'
        >>> if os.path.exists(testdir):
        ...     result = run(*['rm', '--force', '--recursive', testdir])
        >>> result = run(*['mkdir', testdir])

        >>> dir1 = os.path.join(testdir, 'dir1')
        >>> result = run(*['mkdir', dir1])
        >>> assert os.path.exists(dir1)
        >>> dir2 = os.path.join(testdir, 'dir2')
        >>> result = run(*['mkdir', dir2])
        >>> assert os.path.exists(dir2)

        >>> file1 = os.path.join(dir1, 'file1')
        >>> result = run(*['touch', file1])
        >>> file2 = os.path.join(dir1, 'file2')
        >>> result = run(*['touch', file2])

        >>> from denova.os.user import whoami
        >>> if whoami() == 'root':
        ...     copy(dir1, dir2)
        ...     dir1_in_2 = os.path.join(dir2, os.path.basename(dir1))
        ...     assert os.path.isdir(dir1_in_2)
        ...     file1_in_dir1_in_2 = os.path.join(dir1_in_2, 'file1')
        ...     assert os.path.exists(file1_in_dir1_in_2)

        ...     dir3 = os.path.join(testdir, 'subdir', 'dir1')
        ...     result = makedir(dir3)
        ...     copy(dir1, dir3)
        ...     dir1_not_in_3 = os.path.join(dir3, os.path.basename(dir1))
        ...     assert not os.path.isdir(dir1_not_in_3)
        ...     file1_in_dir3 = os.path.join(dir3, 'file1')
        ...     assert os.path.exists(file1_in_dir3)
    '''

    # log.debug('copy(source={}, dest={}, symlinks=={}, ignore=={}, owner=={}, group=={}, perms=={})'.
    #    format(source, dest, symlinks, ignore, owner, group, perms))

    # source must exist, but a dangling link is acceptable
    if not os.path.exists(source) and not os.path.islink(source):
        raise ValueError(f'source "{source}" does not exist')

    if symlinks and os.path.islink(source):
        dest = os.path.join(dest, os.path.basename(source))
        if os.path.exists(dest):
            # log.debug('copy() source is link but dest exists, remove dest: {}'.format(dest))
            os.remove(dest)
        os.symlink(os.readlink(source), dest)

    else:

        if os.path.isdir(dest):
            # append the source's basename to dest if the source
            # and destination basenames are different
            # leave an explicit "basename/basename" dest alone.
            if os.path.basename(dest) != os.path.basename(os.path.dirname(dest)):
                if os.path.basename(source) != os.path.basename(dest):
                    dest = os.path.join(dest, os.path.basename(source))
                    log.debug(f'copy() set path to include source basename: {dest}')

        if os.path.exists(dest):
            log.debug(f'copy() remove dest: {dest}')
            run(*['rm', dest, '--force', '--recursive'])

        if os.path.isdir(source):

            unmount_all(source)

            if (symlinks is True and ignore is None):
                # log.debug('copy() cp({}, {}, archive=true)'.format(source, dest))
                run(*['cp', source, dest, '--archive'])
                set_attributes(dest, owner, group, perms, recursive=True)

            else:
                # log.debug('copy(..., symlinks={}, ignore={})'.format(symlinks, ignore))
                makedir(dest)
                # copystat does *not* affect owner and group
                shutil.copystat(source, dest)
                set_attributes(dest, owner=owner, group=group, recursive=True)

                files = os.listdir(source)
                if ignore is None:
                    ignored = []
                else:
                    ignored = ignore(source, files)
                    # log.debug('copy() ignored set to {}'.format(ignored))

                for file in files:

                    source_path = os.path.join(source, file)
                    if source_path not in ignored:
                        # log.debug('copy() recursing')
                        copy(source_path, dest,
                             symlinks=symlinks, ignore=ignore, owner=owner, group=group, perms=perms)

        else:
            run(*['cp', source, dest, '--preserve'])
            set_attributes(dest, owner, group, perms, recursive=True)

    log.debug('after copy: dest={}, owner=={}, group=={}, perms=={})'. # DEBUG
        format(dest, getuid(dest), getgid(dest), oct(getmode(dest)))) # DEBUG

def merge(source, dest, symlinks=True, force=True, ignore=None, owner=None, group=None, perms=None):
    ''' Merge source to dest dir.

        Like shutil.copytree(), but merges with existing dest dir.
        Unlike shutil.copytree(), symlinks and force default to True.
        This behavior is compatible with denova.os.fs.copy(), and seems to
        almost always be what we want.

        Set ownership and permissions from keywords, or if None copy from
        the corresponding source file.
    '''

    ''' Largely copied from http://code.activestate.com/lists/python-list/191783/
        which in turn is largely copied from shutil.copytree(). '''

    def remove_dest(destname):
        if os.path.lexists(destname) and force:
            log.debug(f'removing {destname}')
            os.remove(destname)
            # assert not os.path.exists(destname) # DEBUG

    if not os.path.exists(source):
        raise ValueError(f'source "{source}" does not exist')

    if not os.path.isdir(dest):
        raise ValueError(f'dest "{dest}" must be a dir')

    # log.debug('merge(source={}, dest={}, symlinks=={}, force=={}, ignore=={}, owner=={}, group=={}, perms=={})'.
    #     format(source, dest, symlinks, force, ignore, owner, group, perms))

    names = os.listdir(source)
    if ignore is not None:
        ignored_names = ignore(source, names)
    else:
        ignored_names = set()

    errors = []

    for name in names:
        if name not in ignored_names:

            sourcename = os.path.join(source, name)
            destname = os.path.join(dest, name)

            try:
                if symlinks and os.path.islink(sourcename):
                    linkto = os.readlink(sourcename)
                    # log.debug('merge() symlink({}, {})'.format(linkto, destname)) #DEBUG
                    remove_dest(destname)
                    os.symlink(linkto, destname)

                elif os.path.isdir(sourcename):
                    if not os.path.isdir(destname):
                        mode = getmode(sourcename)
                        # log.debug('merge() makedirs({}, {})'.format(destname, mode)) #DEBUG
                        makedir(destname, perms=mode)
                    merge(sourcename, destname,
                        symlinks=symlinks, ignore=ignore, force=force,
                        owner=owner, group=group, perms=perms)

                else:
                    # log.debug('merge() copy2({}, {})'.format(sourcename, destname)) #DEBUG
                    remove_dest(destname)
                    shutil.copy2(sourcename, destname)
                    set_attributes(destname, owner=owner, group=group, recursive=True)

                    mode = perms or getmode(sourcename)
                    #log.debug('  src mode: {}'.format(oct(getmode(sourcename)))) #DEBUG
                    chmod(mode, destname)
                # XXX What about devices, sockets etc.?

            except (IOError, os.error) as why:
                log.debug(f'could not merge: {name}')
                errors.append((sourcename, destname, str(why)))

            # catch the shutil.Error from the recursive merge so that we can
            # continue with other files
            except shutil.Error as err:
                log.debug(f'could not merge: {name}')
                errors.extend((sourcename, destname, err.args[0]))

    if errors:
        raise shutil.Error(errors)

def move(source, dest, owner=None, group=None, perms=None):
    ''' Move source to dest.

        move() tries to generally follow the behavior of the
        'mv --force' command.

        move() creates any missing parent dirs of dest.

        If dest is a dir, source is copied into dest.

        Otherwise, source will overwrite any existing dest.

        >>> import os.path, tempfile

        >>> def temp_filename(dir=None):
        ...     if dir is None:
        ...         dir = test_dir
        ...     handle, path = tempfile.mkstemp(dir=dir)
        ...     # we don't need the handle
        ...     os.close(handle)
        ...     return path
        >>>
        >>> test_dir = tempfile.mkdtemp()
        >>>
        >>> # test move to dir
        >>> source = temp_filename()
        >>> dest_dir = tempfile.mkdtemp(dir=test_dir)
        >>> move(source, dest_dir)
        >>> assert not os.path.exists(source)
        >>> basename = os.path.basename(source)
        >>> assert os.path.exists(os.path.join(dest_dir, basename))
        >>>
        >>> # test move to new filename
        >>> source = temp_filename()
        >>> dest_filename = temp_filename(dir=test_dir)
        >>> move(source, dest_filename)
        >>>
        >>> # test move to existing filename
        >>> DATA = 'data'
        >>> source = temp_filename()
        >>> dest_filename = temp_filename()
        >>> with open(source, 'w') as sourcefile:
        ...     sourcefile.write(DATA)
        4
        >>> move(source, dest_filename)
        >>> assert not os.path.exists(source)
        >>> assert os.path.exists(dest_filename)
        >>> with open(dest_filename) as destfile:
        ...     assert DATA == destfile.read()
        >>>
        >>> # clean up
        >>> result = run(*['rm', '--force', '--recursive', test_dir])
        >>> assert result.returncode == 0
    '''

    parent_dir = os.path.dirname(dest)
    if not os.path.exists(parent_dir):
        makedir(parent_dir, owner=owner, group=None, perms=None)
    run(*['mv', '--force', source, dest])
    set_attributes(dest, owner, group, perms, recursive=True)

def clonedirs(sourceroot, destroot, destdir, owner=None, group=None, perms=None):
    ''' Make destdir within destroot. Make any needed intermediate-level
        directories.

        Any mountpoints in sourceroot are unmounted.

        The destdir is relative, and so can not start with '/'.

        Set ownership and permissions from keywords, or if None copy from
        the corresponding dir in sourceroot.

        Only directories are cloned, not data files.

        >>> testroot = '/tmp/test-clonedirs'
        >>> sourceroot = os.path.join(testroot, '1')
        >>> destroot = os.path.join(testroot, '2')
        >>> destdir = 'x/y/z'

        >>> if os.path.exists(sourceroot):
        ...     shutil.rmtree(sourceroot)
        >>> if os.path.exists(destroot):
        ...     shutil.rmtree(destroot)
        >>> assert not os.path.exists(sourceroot)
        >>> assert not os.path.exists(destroot)

        >>> sourcepath = os.path.join(sourceroot, destdir)
        >>> makedir(sourcepath)

        >>> makedir(destroot)

        >>> from denova.os.user import whoami
        >>> destpath = os.path.join(destroot, destdir)
        >>> assert not os.path.exists(destpath)
        >>> if whoami == 'root':
        ...     clonedirs(sourceroot, destroot, destdir)
        ...     os.path.isdir(destpath)
        ... else:
        ...     print(True)
        True
    '''

    if not os.path.exists(sourceroot):
        raise ValueError(f'dir does not exist: {sourceroot}')
    if not os.path.exists(destroot):
        raise ValueError(f'dir does not exist: {destroot}')
    if destdir.startswith('/'):
        raise ValueError(
            f'dest dir must be relative. It can not start with /: {destdir}')

    sourcepath = os.path.join(sourceroot, destdir)
    destpath = os.path.join(destroot, destdir)

    log.debug(f'clonedirs sourcepath={sourcepath}, destpath={destpath}')
    if type(perms) is int:
        log.debug(f'   owner={owner}, group={group}, perms={oct(perms)}')
    else:
        log.debug(f'   owner={owner}, group={group}, perms={perms}')

    # a chroot might have active mounts
    unmount_all(sourcepath)

    if os.path.exists(destpath):
        # if it's a dir, we're done
        if os.path.isfile(destpath):
            raise ValueError(f'dest dir is not a dir: {destdir}')

    else:
        # make any needed parent dirs
        parent = os.path.dirname(destdir)
        clonedirs(sourceroot, destroot, parent, owner, group, perms)

        if perms is None:
            perms = getmode(sourcepath)
        makedir(destpath, owner, group, perms)

def remove(path):
    ''' Remove the path.

        If path is a dir, empty it first by rsyncing an empty dir to the
        path. With a large directory this is much faster than rm or
        shutil.rmtree().

        It is not an error if the path does not exist.

        If path is a link, only remove the link, not the target.

        If path is a mount, raise ValueError.
    '''

    global empty_dir

    if os.path.exists(path):

        if empty_dir is None:
            empty_dir = tempfile.mkdtemp(prefix='denova.os.fs.empty.')

        if os.path.ismount(path):
            raise ValueError(f'path is mount point: {path}')

        elif os.path.islink(path) or os.path.isfile(path):
            run(*['rm', '--force', path])

        else:
            assert os.path.isdir(path)

            run(*['rm', '--force', '--recursive', path])

    assert not os.path.exists(path), f'could not remove {path}'

def relative_path(path):
    ''' Return path as relative path, with no leading or trailing '/'. '''

    return path.strip('/')

def filemode(st_mode):
    ''' Convert integer file permissions to an octal string.

        Function named after python 3.3 os.stat.filemode().

        See python - How to convert a stat output to a unix permissions string - Stack Overflow
            http://stackoverflow.com/questions/17809386/how-to-convert-a-stat-output-to-a-unix-permissions-string

        >>> import os, stat, tempfile

        >>> fd, filename = tempfile.mkstemp()
        >>> os.close(fd)

        >>> # rwx for user and group
        >>> os.chmod(filename, stat.S_IRWXU | stat.S_IRWXG)

        >>> filemode(os.stat(filename).st_mode)
        '-rwxrwx---'

        >>> os.remove(filename)
    '''

    try:
        if S_ISDIR(st_mode):
            is_dir = 'd'
        else:
            is_dir = '-'
    except:   # 'bare except' because it catches more than "except Exception"
        is_dir = '-'
    octal_to_readable = {
        '7':'rwx',
        '6' :'rw-',
        '5' : 'r-x',
        '4':'r--',
        '3':'-wx',
        '2':'-w-',
        '1':'--x',
        '0': '---'}
    perm = str(oct(st_mode)[-3:])
    return is_dir + ''.join(octal_to_readable.get(x,x) for x in perm)

@contextmanager
def restore_file(filename):
    ''' Context manager restores a file to its previous state.

        If the file exists on entry, it is backed up and restored.

        If the file does not exist on entry and does exists on exit,
        it is deleted.
    '''

    exists = os.path.exists(filename)

    if exists:
        # we just want the pathname, not the handle
        # tiny chance of race if someone else gets the temp filename
        handle, backup = tempfile.mkstemp()
        os.close(handle)
        run(*['cp', '--archive', filename, backup])

    try:
        yield

    finally:
        if os.path.exists(filename):
            run(*['rm', filename])
        if exists:
            # restore to original state
            run(*['mv', backup, filename])

def edit_file_in_place(filename, replacements, regexp=False, lines=False):
    ''' Replace text in file.

        'replacements' is a dict of {old: new, ...}.
        Every occurence of each old string is replaced with the
        matching new string.

        If regexp=True, the old string is a regular expression.
        If lines=True, each line is matched separately.

        Perserves permissions.

        >>> # note double backslashes because this is a string within a docstring
        >>> text = (
        ...     'browser.search.defaultenginename=Startpage HTTPS\\n' +
        ...     'browser.search.selectedEngine=Startpage HTTPS\\n' +
        ...     'browser.startup.homepage=https://tails.boum.org/news/\\n' +
        ...     'spellchecker.dictionary=en_US')

        >>> f = tempfile.NamedTemporaryFile(mode='w', delete=False)
        >>> f.write(text)
        178
        >>> f.close()

        >>> HOMEPAGE = 'http://127.0.0.1/'
        >>> replacements = {
        ...     'browser.startup.homepage=.*':
        ...         'browser.startup.homepage={}'.format(HOMEPAGE),
        ...     }

        >>> edit_file_in_place(f.name, replacements, regexp=True, lines=True)

        >>> with open(f.name) as textfile:
        ...     newtext = textfile.read()
        >>> assert HOMEPAGE in newtext

        >>> os.remove(f.name)
    '''

    # import delayed to avoid infinite recursion
    import denova.python.utils

    log.debug(f'edit file in place: {filename}, replacements {replacements}')

    # sometimes replace_strings() gets a type error
    for old, new in replacements.items():
        assert is_string(old), f'replacement old "{old}" should be string but is type {type(old)}'
        assert is_string(new), f'replacement new "{new}" should be string but is type {type(new)}'

    # read text
    mode = os.stat(filename).st_mode
    with open(filename) as textfile:
        text = textfile.read()

    if lines:
        newtext = []
        for line in text.split('\n'):
            # sometimes replace_strings() gets a type error
            assert is_string(line), f'line should be string but is {type(line)}'
            newline = denova.python.utils.replace_strings(line, replacements, regexp)
            newtext.append(newline)
        text = '\n'.join(newtext)
    else:
        text = denova.python.utils.replace_strings(text, replacements, regexp)

    # write text
    with open(filename, 'w') as textfile:
        textfile.write(text)
    os.chmod(filename, mode)
    assert mode == os.stat(filename).st_mode

def why_file_permission_denied(pathname, mode='r'):
    ''' Return string saying why file access didn't work for the current user.

        Checks parent directories.

        If access mode is allowed, returns None.

        mode::
            r       read
            w       write
            x       execute
            s       search
            a,+     append

        The default mode is 'r'. 's' is treated as 'x'. '+' is treated as 'w'.
    '''

    if type(mode) == int:
        mode = filemode(mode)

    reason = None
    while pathname and reason is None:

        for perm in mode:

            if ((perm == 'r' and not is_readable(pathname)) or
                ((perm == 'w' or perm == 'a' or perm == '+') and not is_writeable(pathname)) or
                ((perm == 'x' or perm == 's') and not is_executable(pathname))):

                reason = f'no "{perm}" access for {pathname}'

        if pathname:
            # remove last component of pathname
            parts = pathname.split('/')
            pathname = '/'.join(parts[:-1])

    if reason:
        # import delayed to avoid infinite recursion
        from denova.os.user import whoami

        reason += f' as user {whoami()}'

    return reason

def replace_file(filename, content):
    ''' Replace or create file content.

        Perserves permissions.
    '''

    if os.path.exists(filename):
        mode = os.stat(filename).st_mode
    else:
        mode = None
    with open(filename, 'w') as f:
        f.write(content)
    if mode is not None:
        os.chmod(filename, mode)
        assert mode == os.stat(filename).st_mode

def is_readable(path):
    ''' Return whether the path is readable by the current user '''

    return os.access(path, os.R_OK)

def is_writeable(path):
    ''' Return whether the path is writeable by the current user '''

    return os.access(path, os.W_OK)

def is_executable(path):
    ''' Return whether the path is executable, or if a dir searchable,
        by the current user '''

    return os.access(path, os.X_OK)

def flush(f):
    ''' Flush all file buffers to disk.

        'f' is a file object. The file object must have a fileno() function.
    '''

    # see python docs for os.fsync(fd)
    f.flush()
    os.fsync(f.fileno())

def get_dir_size(start_path = '.'):
    ''' Return size of dir including subdirs. '''

    # from Calculating a directory's size using Python? - Stack Overflow
    #      https://stackoverflow.com/questions/1392413/calculating-a-directorys-size-using-python

    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    return total_size


# moved some code for making a boot drive to file_sys.py

if __name__ == "__main__":
    # import delayed to avoid infinite recursion
    from denova.os.user import whoami

    if whoami == 'root':
        import doctest
        doctest.testmod()
    else:
        print("fs's doctests must be run as root")



'''
    User utilities.

    Copyright 2010-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

from contextlib import contextmanager
import grp
import os
import pwd
import sys

from denova.os.command import run
from denova.python.log import get_log


log = get_log()

def whoami():
    ''' Get name of current user.

        Calling whoami() writes to /var/log/auth.log. To avoid flooding that
        log, try to cache the result from whoami() when you can.

        >>> assert whoami() == pwd.getpwuid(os.geteuid()).pw_name
    '''

    who = run('whoami').stdout
    return who

def require_user(user):
    ''' Require a specific current user. '''

    current_user = whoami()
    if current_user != user:
        # import this late to avoid conflicts
        from denova.python.utils import stacktrace
        msg = f'This program must be run as {user}. Current user is {current_user}.'
        log.debug(msg)
        log.debug(stacktrace())
        sys.exit(msg)

def su(newuser, set_home_dir=True):
    ''' Login as newuser.

        Use sudo() if you want to return to the original user later.
        This will usually only work if you call sudo() as root.

        This function only successfully changes the euid, not the uid.
        Programs which use the uid, such as ssh and gpg, need extra help.
        One solution is to prefix commands with "sudo -u USER".
        See the reinhardt.ssh module.

        This function is unreliable with ssh, os.openpty(), and more.
        A workaround is to use sudo in an enclosing bash script. Or::

            # if not user VM_OWNER then relaunch this program
            if whoami().strip() != VM_OWNER:
                os.execvp( 'sudo' , ['sudo', '-u', VM_OWNER] + sys.argv)

        Raises OsError if user does not exist or current
        user does not has permission to log in as new user. '''

    if whoami() != newuser:
        uid = getuid(newuser)
        os.seteuid(uid)
        # why doesn't this work?
        try:
            # See http://stackoverflow.com/questions/7529252/operation-not-permitted-on-using-os-setuid-python
            # if os.fork():
            #     os._exit(0)

            os.setuid(uid)
        except:   # 'bare except' because it catches more than "except Exception"
            # print(traceback.format_exc().strip()) # DEBUG
            # print('ERROR IGNORED. Because os.setuid() does not appear to work even for root') # DEBUG
            pass

    require_user(newuser)
    if set_home_dir:
        os.environ['HOME'] = getdir(newuser)

def sudo(username=None, set_home_dir=True):
    ''' Context manager to temporarily run code as another user.

        This will usually only work if you call sudo() as root.

        Use su() if you do not want to return to the original user later.
        If you are root, this function only sets the euid.
        Some programs use the uid instead of the euid, such as ssh or gpg.

        The current user must have the NOPASSWD option set in /etc/sudoers.
        Otherwise sudo will hang. (This is true for sh. Is NOPASSWD needed
        for this function?)

        WARNING: This function is unreliable with ssh, os.openpty(), and more.
        A workaround is to use sudo in an enclosing bash script. Or::

            # if not user VM_OWNER then relaunch this program
            if whoami().strip() != VM_OWNER:
                os.execvp( 'sudo' , ['sudo', '-u', VM_OWNER] + sys.argv)

        >>> import denova.os.user

        >>> original_user = denova.os.user.whoami()
        >>> if original_user == 'root':
        ...
        ...     for user in denova.os.user.users():
        ...         if user != original_user:
        ...
        ...             with denova.os.user.sudo(user):
        ...                 assert denova.os.user.whoami() == user, 'could not sudo as {}'.format(user)
        ...             assert denova.os.user.whoami() == original_user
    '''

    @contextmanager
    def null_contextmanager():
        yield

    @contextmanager
    def active_contextmanager():
        """
        # in python 2.7 os.fork() results in
        #     RuntimeError: not holding the import lock
        # apparently python 3 does not have the bug
        # see  http://bugs.python.org/issue18122
        child = os.fork()
        if child:
            try:
                prev_uid = os.getuid()
                uid = getuid(username)
                os.setuid(uid)
                if set_home_dir:
                    os.environ['HOME'] = getdir(username)
                yield
            finally:
                try:
                    os.setuid(prev_uid)
                except:   # 'bare except' because it catches more than "except Exception"
                    pass
                if set_home_dir:
                    os.environ['HOME'] = getdir(prev_user)

        else:
            os.waitpid(child, 0)

        """
        try:
            uid = getuid(username)
            if prev_user == 'root':
                os.seteuid(uid)
            else:
                os.setuid(uid)
            # os.setuid(uid) # DEBUG
            if set_home_dir:
                os.environ['HOME'] = getdir(username)
            yield
        finally:
            prev_uid = getuid(prev_user)
            try:
                os.seteuid(prev_uid)
            except:   # 'bare except' because it catches more than "except Exception"
                pass
            else:
                if set_home_dir:
                    os.environ['HOME'] = getdir(prev_user)

    if not username:
        username = 'root'
        # log.debug('sudo() using default user root')

    prev_user = whoami()
    if username == prev_user:
        # no need to sudo, and avoid spurious "is not in the sudoers file" error
        context = null_contextmanager()
    else:
        context = active_contextmanager()

    return context

def force(user):
    ''' If current user is not 'user', relaunch program as 'user'.

        Example::

            if denova.os.user.whoami() == 'root':
                root_setup()
                ...

                # drop privs; relaunch program as USER
                denova.os.user.force(USER)

            # continue as USER
            assert whoami() == USER

    '''

    if whoami() != user:

        # import this late to avoid conflicts
        #from denova.os.fs import is_executable
        #this_program = sys.argv[0]
        # we need something like this, but this doesn't work
        # with sudo(user):
        #     assert is_executable(this_program), '{} must be executable as {}'.format(this_program, user)

        for f in [sys.stdout, sys.stderr]:
            f.flush()
            try:
                os.fsync(f.fileno())
            except OSError:
                # stdout and stderr usually aren't' real files
                pass
        os.execvp( 'sudo' , ['sudo', '-u', user] + sys.argv)

def users():
    '''
        Get active users.

        >>> len(users()) > 0
        True

    '''

    return set(run('users').stdout.split())

def getuid(username):
    ''' Return uid for username. '''

    name_info = pwd.getpwnam(username)
    return name_info.pw_uid

def getuid_name(uid):
    ''' Return user name for uid.

        >>> import pwd

        >>> getuid_name('not an int')
        Traceback (most recent call last):
            ...
        ValueError: uid is not an int: not an int

        >>> for entry in pwd.getpwall():
        ...     assert getuid_name(entry.pw_uid) == entry.pw_name
    '''

    try:
        uid = int(uid)
    except ValueError:
        raise ValueError(f'uid is not an int: {uid}')

    name = None
    for entry in pwd.getpwall():
        if entry.pw_uid == uid:
            name = entry.pw_name

    if name is None:
        raise ValueError(f'Not a valid uid {uid}')

    return name

def getgid(groupname):
    ''' Return gid for groupname. '''

    name_info = grp.getgrnam(groupname)
    return name_info.gr_gid

def getgid_name(gid):
    ''' Return group name for gid.

        >>> import grp

        >>> getgid_name('string')
        Traceback (most recent call last):
            ...
        ValueError: gid is not an int: string

        >>> for entry in grp.getgrall():
        ...     assert getgid_name(entry.gr_gid) == entry.gr_name
    '''

    try:
        gid = int(gid)
    except ValueError:
        raise ValueError(f'gid is not an int: {gid}')

    name = None
    for entry in grp.getgrall():
        if entry.gr_gid == gid:
            name = entry.gr_name

    if name is None:
        raise ValueError(f'Not a valid gid {gid}')

    return name

def getdir(username=None):
    ''' Return home dir for username. '''

    if username is None:
        username = whoami()

    name_info = pwd.getpwnam(username)
    return name_info.pw_dir


if __name__ == "__main__":

    import doctest
    doctest.testmod()



'''
    Log Django debug pages.

    Copyright 2010-2020 DeNova
    Last modified: 2020-10-22

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os
from tempfile import NamedTemporaryFile

try:
    from django.utils.deprecation import MiddlewareMixin
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.django_addons.utils import is_django_error_page
from denova.python.log import get_log

log = get_log()


class DebugMiddleware(MiddlewareMixin):
    ''' Write to debugging log.

        Logs Django debug pages and says it's an error. '''

    def process_response(self, request, response):

        def logit(why):
            log(why)
            log(f'response: {response!r}')
            log(f'request: {request!r}')

        try:
            if is_django_error_page(response.content):
                with NamedTemporaryFile(
                    prefix='django.debug.page.', suffix='.html',
                    delete=False) as htmlfile:
                    htmlfile.write(response.content)
                os.chmod(htmlfile.name, 0o644)
                log(f'django app error: django debug page at {htmlfile.name}')
            elif response.status_code >= 400:
                logit(f'http error {response.status_code:d}')
        except AttributeError as ae:
            log(f'ignored in denova.django_addons.middleware.DebugMiddleware.process_response(): {ae}')

        return response



'''
    Text formating.

    Because this module uses some modules that use this one,
    imports that are not from standard libs should not be
    at global scope. Put the import where it's used.

    Copyright 2008-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import json
import locale
import os
import pprint
import re
from traceback import format_exc

UNICODE_DECODINGS = ['utf-8',
                     'iso-8859-14',
                     locale.getpreferredencoding(False)]
DETAILED_LOG = '/tmp/denova.python.format.detailed.log'

# delayed import of log so denova.python.log can use this module
_log = None


def log(message):
    global _log
    if not _log:
        from denova.python.log import get_log
        _log = get_log()
    _log(message)


# delayed open of detailed log so other users can use this module
_detailed_log = None


def detailed_log(message):
    global _detailed_log
    if not _detailed_log:
        _detailed_log = open(DETAILED_LOG, 'w')
        # user frequently changes, so loosen access
        os.chmod(DETAILED_LOG, 0o660)
    _detailed_log.write(message)
    _detailed_log.flush()

def pretty(obj, indent=4, base_indent=0):
    ''' Prettyprint 'pprint' replacement.

        Places every dictionary item on a separate line in key order.
        Formats nested dictionaries.

        For long lists, places every item on a separate line.

        'indent' is the increment for each indentation level, and defaults to 4.
        'base_indent' is the current indentation, and defaults to 0.

    >>> import datetime
        >>> data = {
        ...     'a': 1,
        ...     'c': 2,
        ...     'b': 'hi',
        ...     'x': {1: 'a', 2: 'b'},
        ...     'e': datetime.timedelta(days=9, seconds=11045),
        ...     'd': 'ho',
        ...     'g': datetime.datetime(2000, 1, 2, 3, 4, 5, 6),
        ...     'f': datetime.date(2000, 1, 2),
        ...     'h': datetime.time(1, 2, 3, 4),
        ...     }
        >>> data['l'] = [data['a'], data['b'], data['c'], data['d'], data['e'], data['f'], data['g'], data['h'], data['x']]
        >>> p = pretty(
        ...     data,
        ...     indent=4
        ...     )
        >>> print(p)
        {
            'a': 1,
            'b': 'hi',
            'c': 2,
            'd': 'ho',
            'e': datetime.timedelta(days=9, seconds=11045),
            'f': datetime.date(2000, 1, 2),
            'g': datetime.datetime(2000, 1, 2, 3, 4, 5, 6),
            'h': datetime.time(1, 2, 3, 4),
            'l': [
                1,
                'hi',
                2,
                'ho',
                datetime.timedelta(days=9, seconds=11045),
                datetime.date(2000, 1, 2),
                datetime.datetime(2000, 1, 2, 3, 4, 5, 6),
                datetime.time(1, 2, 3, 4),
                {
                    1: 'a',
                    2: 'b',
                },
            ],
            'x': {
                1: 'a',
                2: 'b',
            },
        }

        >>> p1 = eval(p)
        >>> print(pretty(p1, indent=4))
        {
            'a': 1,
            'b': 'hi',
            'c': 2,
            'd': 'ho',
            'e': datetime.timedelta(days=9, seconds=11045),
            'f': datetime.date(2000, 1, 2),
            'g': datetime.datetime(2000, 1, 2, 3, 4, 5, 6),
            'h': datetime.time(1, 2, 3, 4),
            'l': [
                1,
                'hi',
                2,
                'ho',
                datetime.timedelta(days=9, seconds=11045),
                datetime.date(2000, 1, 2),
                datetime.datetime(2000, 1, 2, 3, 4, 5, 6),
                datetime.time(1, 2, 3, 4),
                {
                    1: 'a',
                    2: 'b',
                },
            ],
            'x': {
                1: 'a',
                2: 'b',
            },
        }
    '''

    max_list_width = 60

    if isinstance(obj, dict):
        p = '{\n'
        base_indent += indent
        try:
            keys = sorted(obj.keys())
        except:   # 'bare except' because it catches more than "except Exception"
            keys = obj.keys()
        for key in keys:
            p += (' ' * base_indent) + repr(key) + ': '
            value = obj[key]
            p += pretty(value, indent=indent, base_indent=base_indent)
            p += ',\n'
        base_indent -= indent
        p += (' ' * base_indent) + '}'

    elif isinstance(obj, list):
        p = '[\n'
        base_indent += indent
        for item in obj:
            p += (' ' * base_indent) + pretty(item, indent=indent, base_indent=base_indent)
            p += ',\n'
        base_indent -= indent
        p += (' ' * base_indent) + ']'
        # put short lists on one line
        if len(p) < max_list_width:
            p = p.replace('\n ', ' ')
            p = p.replace('  ', ' ')

    else:
        pp = pprint.PrettyPrinter(indent=indent)
        try:
            p = pp.pformat(obj)
        except:   # 'bare except' because it catches more than "except Exception"
            try:
                log(f'unable to pretty print object: {obj}')
                p = repr(obj)
                log('object len: %d' % len('{}'.format(p)))
            except:   # 'bare except' because it catches more than "except Exception"
                log(format_exc())
                from denova.python.utils import last_exception_only
                p = f'denova.python.format.pretty ERROR:{last_exception_only()}'

    return p

def add_commas(number):
    ''' Add commas to a number,

        >>> print(add_commas(0))
        0
        >>> print(add_commas(1))
        1
        >>> print(add_commas(1234))
        1,234
        >>> print(add_commas(1234567))
        1,234,567
        >>> print(add_commas(0.0))
        0.0
        >>> print(add_commas(0.1234))
        0.1234
        >>> print(add_commas(1.1234))
        1.1234
        >>> print(add_commas(1234.1234))
        1,234.1234
        >>> print(add_commas(1234567.1234))
        1,234,567.1234
    '''

    if '.' in str(number):
        (integer, decimal) = str(number).split('.')
    else:
        integer = str(number)
        decimal = ''

    formattedNumber = ''
    while len(integer) > 3:
        formattedNumber = "," + integer[-3:] + formattedNumber
        integer = integer[:-3]
    formattedNumber = integer + formattedNumber

    if decimal:
        formattedNumber += '.' + decimal

    return formattedNumber

def s_if_plural(number):
    ''' Return an empty string if the number is one, else return the letter \'s\'.
        This is used to form standard plural nouns.

        >>> print('house' + s_if_plural(0))
        houses
        >>> print('house' + s_if_plural(1))
        house
        >>> print('house' + s_if_plural(2))
        houses

    '''

    if number == 1:
        result = ''
    else:
        result = 's'
    return result

def replace_angle_brackets(s):
    ''' Replace '<' with '&lt;' and '>' with '&gt;'.

        This allows html to display correctly when embedded in html. '''

    s = s.replace('<', '&lt;')
    s = s.replace('>', '&gt;')
    return s

def camel_back(s):
    ''' Combine words into a string with no spaces and every word capitalized.
        Already capitalized letters after the first letter are preserved.

        >>> camel_back('wikipedia article name')
        'WikipediaArticleName'
        >>> camel_back('WikiPedia CamelBack')
        'WikiPediaCamelBack'

        '''

    words = s.split(' ');
    camel_back_words = []
    for word in words:
        # the word may itself be camel back, or at least have some capitalized letters
        camel_back_words.append(word[:1].capitalize() + word[1:])
    return ''.join(camel_back_words)

def pretty_html(html, parser=None):
    ''' Prettyprint html.

        Requires BeautifulSoup, python3-tidylib, or python3-utidylib.

        'parser' specifies the parser used by BeautifulSoup. 'html5lib'
        is the most lenient and the default. 'lxml' is the fastest.

        >>> html = '<head><title>Test HTML</title></head><body>The test text</body>'
        >>> html_prettied = pretty_html(html)
        >>> assert isinstance(html_prettied, str)
        >>> '<html>' in html_prettied
        True
        >>> '</html>' in html_prettied
        True

        >>> bytes_html = b'<head><title>Test HTML</title></head><body>The test text</body>'
        >>> pretty_bytes_html = pretty_html(bytes_html)
        >>> assert isinstance(pretty_bytes_html, bytes)
        >>> b'<html>' in pretty_bytes_html
        True
        >>> b'</html>' in pretty_bytes_html
        True
    '''

    def append_line(line):
        line = line.strip()
        log(f'line: {line}') # DEBUG
        lines.append(line)

    #log('type(html): {}'.format(type(html)))
    #log('repr(html): {}'.format(repr(html)))
    WAS_STRING = isinstance(html, str)
    #log('WAS_STRING: {}'.format(WAS_STRING))
    if WAS_STRING:
        html = to_bytes(html)
        #log('type(html) after to_bytes: {}'.format(type(html)))

    if parser is None:
        parser = 'html5lib'
        log('using html5lib parser')

    # try various html prettyprinters
    p_html = None

    try:
        from bs4 import BeautifulSoup

        log(f'clean html with beautifulsoup prettify, parser: {parser}')

        soup = BeautifulSoup(html, features=parser)
        p_html = soup.prettify(encoding='utf-8')

    except ImportError:
        pass

    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    if not p_html:
        try:
            # python3-tidylib
            from tidylib import tidy_document

            if b'<frameset' in html:
                raise ValueError('tidy_document() does not work with framesets')

            p_html, errors = tidy_document(html)
            if errors:
                # rss is not html
                if '<rss> is not recognized' in errors:
                    log("Warning: tidy prettyprinter can't format rss")
                else:
                    log("Warning: tidy prettyprinter found errors")
                    # log("see {}".format(DETAILED_LOG))
                    # detailed_log('tidy error: {}\n'.format(errors))
            elif not p_html:
                log('tidy returned an empty page')

        except ImportError:
            log('No module named tidylib. Install debian package python3-tidylib or pypi package pytidylib.')
        except:   # 'bare except' because it catches more than "except Exception"
            log(format_exc())

    if not p_html:
        try:
            # python3-utidylib
            import tidy

            options = dict(output_xhtml=1, add_xml_decl=1, indent=1, tidy_mark=0)
            try:
                p_html = str(tidy.parseString(html, **options))
            except AttributeError as aerr:
                # not debugged:
                #      module 'tidy' has no attribute 'parseString'
                # maybe need to install/reinstall python3-utidylib
                log.warning(str(aerr))

        except ImportError:
            log('No module named tidy')
        except:   # 'bare except' because it catches more than "except Exception"
            log(format_exc())

        else:
            if not p_html:
                log('empty string from python3-utidylib')

    if not p_html:
        # import late to avoid conflicts
        from denova.python.utils import stacktrace

        # ad hoc prettyprinter, indents tag/endtag blocks

        log('denova.python.format() NOT WORKING') # DEBUG
        log(stacktrace()) # the caller needs to handle a returned value of None
        return None

        # split into lines
        lines = []
        line = ''
        quote_char = None # or quote char
        for ch in html:

            #log('ch: {}'.format(repr(ch))) # DEBUG
            # if type(html) is bytes, then type(ch) is int
            if isinstance(ch, int):
                ch = chr(ch)
                #log('ch changed to: {}'.format(repr(ch))) # DEBUG

            if quote_char:
                #log('in quote') # DEBUG
                # don't look for tags in quoted strings
                line = line + ch
                if ch == quote_char:
                    quote_char = None

            elif ch in ('"', "'"):
                #log('start quote: {}'.format(ch)) # DEBUG
                quote_char = ch
                line = line + ch

            # look for tags
            elif ch == '<':
                log('start tag, end of previous line') # DEBUG
                if line:
                    lines.append(line.strip())
                line = '<'
            elif ch == '>':
                log('end tag char') # DEBUG
                if line:
                    log('end tag char, so end of line') # DEBUG
                    line = line + ch
                    append_line(line)
                    line = ''
                else:
                    log('probably not a tag') # DEBUG
                    line = line + ch

            else:
                line = line + ch

        # finish last line, if any
        if line:
            append_line(line)
        #log('lines done')

        # indent blocks with endtags
        start_tag_pattern = re.compile(b'^<\s*([A-Za-z]+)\b.*>')
        end_tag_pattern = re.compile(b'^</\s*([A-Za-z]+)\b.*>')
        reversed_rough = []
        tags = []
        indent = 0
        for line in reversed(lines):

            end_match = end_tag_pattern.match(line)
            if end_match:
                endtag = end_match.group(1)
                tags.append(endtag)
                reversed_rough.append((indent * b'    ') + line)
                indent += 1

            else:
                start_match = start_tag_pattern.match(line)
                if start_match:
                    starttag = start_match.group(1)
                    if starttag in tags:
                        del tags[tags.index(starttag)]
                        if indent:
                            indent -= 1
                reversed_rough.append((indent * b'    ') + line)

        # remove excess indent
        first = reversed_rough[-1]
        excess_count = len(first) - len(first.lstrip())
        excess = b' ' * excess_count
        reversed_lines = []
        for line in reversed_rough:
            if line.startswith(excess):
                line = line[excess_count:]
            reversed_lines.append(line)

        # join lines
        lines = reversed(reversed_lines)
        p_html = b'\n'.join(lines)

    assert p_html, 'Unable to prettyprint. Tried BeautifulSoup, python3-tidylib, python3-utidylib, and ad hoc'
    #log('p_html:\n{}'.format(p_html)) # DEBUG

    #log('at end of p_html() WAS_STRING: {}'.format(WAS_STRING)) # DEBUG
    if WAS_STRING:
        #log('at end of p_html() call to_string(html)') # DEBUG
        p_html = to_string(p_html)
    #log('at end of p_html() type(html): {}'.formattype(html))) # DEBUG

    return p_html

def pretty_json(json_string, indent=4):
    ''' Prettyprint json string.

        >>> json_string = '{"b": 2, "a": 1, "d": {"y": "b", "x": "a"}, "c": 3}'
        >>> print(pretty_json(json_string))
        {
            "a": 1,
            "b": 2,
            "c": 3,
            "d": {
                "x": "a",
                "y": "b"
            }
        }

    '''

    decoded = json.loads(json_string)
    encoded = json.dumps(decoded, indent=indent, sort_keys=True)
    return encoded

def pretty_called_process_error(cpe):
    ''' Pretty-print subprocess.CalledProcessError '''

    def pretty_text(label, text):
        if text:
            text = text.decode().strip()
            text.replace('\n', '\n\t\t')
            text = '\t' + label + ':\n\t\t' + text

        else:
            text = ''

        return text

    pretty = 'subprocess.CalledProcessError\n'

    command_args = list(map(str, cpe.args))
    command_str = ' '.join(command_args)
    pretty = pretty + '\t' + command_str + '\n'
    pretty = pretty + '\t' + f'error returncode: {cpe.returncode}' + '\n'
    pretty = pretty + pretty_text('stdout', cpe.stdout)
    pretty = pretty + pretty_text('stderr', cpe.stderr)

    return pretty

def encode_unicode(string):
    ''' DEPRECATED. Use to_bytes().

        Deprecated because no one can remember what .encode()
        and .decode() do.

        Convert string to bytes.
        Replacement for STRING.encode().

        If not decoded raises UnicodeEncodeError.
    '''

    return to_bytes(string)

def decode_unicode(string):
    ''' DEPRECATED. Use to_string().

        Deprecated because no one can remember what .encode()
        and .decode() do.

        Convert bytes to a unicode string.
        Replacement for bytes.decode().
    '''

    return to_string(string)

def read_unicode(stream, errors=None):
    ''' Try to decode a stream of bytes to a unicode string.
        Replacement for stream.read().

        There has got to be a standard way to do this that works, but haven't found it.

        If not decoded raises UnicodeDecodeError.

        If stream is not bytes raises TypeError.

        To do:
            Use encoding hints, e.g. http's Content-Encoding::
                content-type = text/html; charset=UTF-8
                or from python::
                    charset = denova.net.http_addons.content_encoding_charset(params)
                or html's content-type::
                    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    '''

    if errors is None:
        errors = 'strict'

    decodings = UNICODE_DECODINGS
    decoded = None

    data = stream.read()

    if isinstance(data, str):
        decoded = data

    else:
        for encoding in decodings:
            if decoded is None:
                try:
                    decoded = str(data, encoding, errors)

                except UnicodeDecodeError:
                    pass

    if decoded is None:
        raise UnicodeDecodeError(f'unable to decode unicode using {repr(UNICODE_DECODINGS)}')

    return decoded

def to_bytes(obj):
    ''' Convert string to bytes.

        'obj' must be a string or bytes. If obj is bytes or bytearray, it is unchanged.

        Convenience function because no one can remember what .encode()
        and .decode() do.

        Replacement for string.encode().
    '''

    if isinstance(obj, (bytearray, bytes)):
        encoded = obj

    else:
        if repr(obj).startswith("b'"): raise Exception(f'apparent byte literal as string: {obj}') # DEBUG
        encoded = None
        for encoding in UNICODE_DECODINGS:
            if encoded is None:
                try:
                    encoded = obj.encode(encoding)
                    if isinstance(encoded, str):
                        log(f'unable to encode string. type after encoding is {type(encoded)}')
                        encoded = None

                except UnicodeEncodeError:
                    pass

        if encoded is None:
            raise UnicodeEncodeError(f'unable to encode unicode using {repr(UNICODE_DECODINGS)}')

        assert isinstance(encoded, bytes)

    return encoded

def to_string(obj):
    ''' Convert bytes to string.

        'obj' must be a string or bytes. If obj is a string, it is unchanged.

        If not decoded raises UnicodeDecodeError.

        Replacement for bytes.decode().
    '''

    if isinstance(obj, str):
        decoded = obj

    else:
        decoded = None
        for encoding in UNICODE_DECODINGS:
            if decoded is None:
                try:
                    decoded = obj.decode(encoding)

                except UnicodeDecodeError:
                    pass

        if decoded is None:
            raise UnicodeDecodeError(f'unable to decode unicode using {repr(UNICODE_DECODINGS)}')

        assert isinstance(decoded, str)

    return decoded


if __name__ == "__main__":
    import doctest
    doctest.testmod()
